#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the 'amsr2_l2_gaasp' reader."""
from __future__ import annotations

import datetime as dt
import os
from pathlib import Path

import dask.array as da
import numpy as np
import pytest
import xarray as xr

MBT_FILENAME = "AMSR2-MBT_v2r2_GW1_s202008120558310_e202008120607010_c202008120637340.nc"
PRECIP_FILENAME = "AMSR2-PRECIP_v2r2_GW1_s202008120558310_e202008120607010_c202008120637340.nc"
OCEAN_FILENAME = "AMSR2-OCEAN_v2r2_GW1_s202008120558310_e202008120607010_c202008120637340.nc"
SEAICE_NH_FILENAME = "AMSR2-SEAICE-NH_v2r2_GW1_s202008120558310_e202008120607010_c202008120637340.nc"
SEAICE_SH_FILENAME = "AMSR2-SEAICE-SH_v2r2_GW1_s202008120558310_e202008120607010_c202008120637340.nc"
SNOW_FILENAME = "AMSR2-SNOW_v2r2_GW1_s202008120558310_e202008120607010_c202008120637340.nc"
SOIL_FILENAME = "AMSR2-SOIL_v2r2_GW1_s202008120558310_e202008120607010_c202008120637340.nc"

EXAMPLE_FILENAMES = [
    MBT_FILENAME,
    PRECIP_FILENAME,
    OCEAN_FILENAME,
    SEAICE_NH_FILENAME,
    SEAICE_SH_FILENAME,
    SNOW_FILENAME,
    SOIL_FILENAME,
]


def _get_shared_global_attrs(filename):
    attrs = {
        "time_coverage_start": "2020-08-12T05:58:31.0Z",
        "time_coverage_end": "2020-08-12T06:07:01.0Z",
        "platform_name": "GCOM-W1",
        "instrument_name": "AMSR2",
    }
    return attrs


def _create_two_res_gaasp_dataset(filename):
    """Represent files with two resolution of variables in them (ex. OCEAN)."""
    lon_var_hi = xr.DataArray(
        da.zeros((10, 10), dtype=np.float32),
        dims=("Number_of_Scans", "Number_of_hi_rez_FOVs"),
        attrs={"standard_name": "longitude"},
    )
    lat_var_hi = xr.DataArray(
        da.zeros((10, 10), dtype=np.float32),
        dims=("Number_of_Scans", "Number_of_hi_rez_FOVs"),
        attrs={"standard_name": "latitude"},
    )
    lon_var_lo = xr.DataArray(
        da.zeros((10, 10), dtype=np.float32),
        dims=("Number_of_Scans", "Number_of_low_rez_FOVs"),
        attrs={"standard_name": "longitude"},
    )
    lat_var_lo = xr.DataArray(
        da.zeros((10, 10), dtype=np.float32),
        dims=("Number_of_Scans", "Number_of_low_rez_FOVs"),
        attrs={"standard_name": "latitude"},
    )
    swath_var1 = xr.DataArray(
        da.zeros((10, 10), dtype=np.float32),
        dims=("Number_of_Scans", "Number_of_hi_rez_FOVs"),
        coords={"some_longitude_hi": lon_var_hi, "some_latitude_hi": lat_var_hi},
        attrs={"_FillValue": -9999.0, "scale_factor": 0.5, "add_offset": 2.0},
    )
    swath_var2 = xr.DataArray(
        da.zeros((10, 10), dtype=np.float32),
        dims=("Number_of_Scans", "Number_of_low_rez_FOVs"),
        coords={"some_longitude_lo": lon_var_lo, "some_latitude_lo": lat_var_lo},
        attrs={"_FillValue": -9999.0},
    )
    swath_int_var = xr.DataArray(
        da.zeros((10, 10), dtype=np.uint16),
        dims=("Number_of_Scans", "Number_of_low_rez_FOVs"),
        attrs={"_FillValue": 100, "comment": "Some comment"},
    )
    not_xy_dim_var = xr.DataArray(da.zeros((10, 5), dtype=np.float32), dims=("Number_of_Scans", "Time_Dimension"))
    time_var = xr.DataArray(da.zeros((5,), dtype=np.float32), dims=("Time_Dimension",))
    ds_vars = {
        "swath_var_hi": swath_var1,
        "swath_var_low": swath_var2,
        "swath_var_low_int": swath_int_var,
        "some_longitude_hi": lon_var_hi,
        "some_latitude_hi": lat_var_hi,
        "some_longitude_lo": lon_var_lo,
        "some_latitude_lo": lat_var_lo,
        "not_xy_dim_var": not_xy_dim_var,
        "time_var": time_var,
    }
    attrs = _get_shared_global_attrs(filename)
    ds = xr.Dataset(ds_vars, attrs=attrs)
    return ds


def _create_gridded_gaasp_dataset(filename):
    """Represent files with gridded products."""
    grid_var = xr.DataArray(
        da.zeros((10, 10), dtype=np.float32),
        dims=("Number_of_Y_Dimension", "Number_of_X_Dimension"),
        attrs={"_FillValue": -9999.0, "scale_factor": 0.5, "add_offset": 2.0},
    )
    latency_var = xr.DataArray(
        da.zeros((10, 10), dtype=np.int32),
        dims=("Number_of_Y_Dimension", "Number_of_X_Dimension"),
        attrs={
            "units": "seconds",
            "_FillValue": -9999,
        },
    )
    time_var = xr.DataArray(da.zeros((5,), dtype=np.float32), dims=("Time_Dimension",))
    ds_vars = {
        "grid_var": grid_var,
        "latency_var": latency_var,
        "time_var": time_var,
    }
    attrs = _get_shared_global_attrs(filename)
    return xr.Dataset(ds_vars, attrs=attrs)


def _create_one_res_gaasp_dataset(filename):
    """Represent files with one resolution of variables in them (ex. SOIL)."""
    lon_var_lo = xr.DataArray(
        da.zeros((10, 10), dtype=np.float32),
        dims=("Number_of_Scans", "Number_of_low_rez_FOVs"),
        attrs={"standard_name": "longitude"},
    )
    lat_var_lo = xr.DataArray(
        da.zeros((10, 10), dtype=np.float32),
        dims=("Number_of_Scans", "Number_of_low_rez_FOVs"),
        attrs={"standard_name": "latitude"},
    )
    swath_var2 = xr.DataArray(
        da.zeros((10, 10), dtype=np.float32),
        dims=("Number_of_Scans", "Number_of_low_rez_FOVs"),
        coords={"some_longitude_lo": lon_var_lo, "some_latitude_lo": lat_var_lo},
        attrs={"_FillValue": -9999.0, "scale_factor": 0.5, "add_offset": 2.0},
    )
    swath_int_var = xr.DataArray(
        da.zeros((10, 10), dtype=np.uint16),
        dims=("Number_of_Scans", "Number_of_low_rez_FOVs"),
        attrs={"_FillValue": 100, "comment": "Some comment"},
    )
    time_var = xr.DataArray(da.zeros((5,), dtype=np.float32), dims=("Time_Dimension",))
    ds_vars = {
        "swath_var": swath_var2,
        "swath_var_int": swath_int_var,
        "some_longitude_lo": lon_var_lo,
        "some_latitude_lo": lat_var_lo,
        "time_var": time_var,
    }
    attrs = _get_shared_global_attrs(filename)
    return xr.Dataset(ds_vars, attrs=attrs)


def _create_fake_gaasp_file(fake_path: Path, **kwargs):
    """Create a Dataset similar to reading an actual file with xarray.open_dataset."""
    filename = fake_path.name
    if filename in [MBT_FILENAME, PRECIP_FILENAME, OCEAN_FILENAME]:
        ds = _create_two_res_gaasp_dataset(filename)
    elif filename in [SEAICE_NH_FILENAME, SEAICE_SH_FILENAME]:
        ds = _create_gridded_gaasp_dataset(filename)
    else:
        ds = _create_one_res_gaasp_dataset(filename)
    ds.to_netcdf(fake_path, )


def _create_fake_ocean_wind_speed_file(filename, **kwargs):
    """Create a Dataset similar to reading an actual OCEAN file with wind speed data."""
    ds = _create_two_res_gaasp_dataset(filename)
    wspd_data = da.arange(10 * 10, dtype=np.float32).reshape((10, 10))
    wspd_data_arr = xr.DataArray(
        wspd_data,
        dims=("Number_of_Scans", "Number_of_low_rez_FOVs"),
        coords={"some_longitude_lo": ds["some_longitude_lo"], "some_latitude_lo": ds["some_latitude_lo"]},
        attrs={"_FillValue": -9999.0},
    )
    ds["WSPD"] = wspd_data_arr
    wspd_qc_data = np.zeros((10, 10), dtype=np.int32).reshape((10, 10))
    wspd_qc_data[-1, -1] = 1
    wspd_qc_data_arr = xr.DataArray(
        da.from_array(wspd_qc_data),
        dims=("Number_of_Scans", "Number_of_low_rez_FOVs"),
        attrs={"_FillValue": -9999, "units": "1"},
    )
    ds["WSPD_QC"] = wspd_qc_data_arr
    ds.to_netcdf(filename)


class TestGAASPReader:
    """Tests for the GAASP reader."""

    yaml_file = "amsr2_l2_gaasp.yaml"

    def setup_method(self):
        """Wrap pygrib to read fake data."""
        from satpy._config import config_search_paths

        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))

    @pytest.mark.parametrize(
        ("filenames", "expected_loadables"),
        [
            (EXAMPLE_FILENAMES, 7),
            ([MBT_FILENAME], 1),
            ([PRECIP_FILENAME], 1),
            ([OCEAN_FILENAME], 1),
            ([SEAICE_NH_FILENAME], 1),
            ([SEAICE_SH_FILENAME], 1),
            ([SNOW_FILENAME], 1),
            ([SOIL_FILENAME], 1),
        ],
    )
    def test_reader_creation(self, tmp_path, filenames, expected_loadables):
        """Test basic initialization."""
        from satpy.readers.core.loading import load_reader

        filenames = [tmp_path / filename for filename in filenames]
        for fn in filenames:
            _create_fake_gaasp_file(fn)
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames(filenames)
        assert len(loadables) == expected_loadables
        r.create_filehandlers(loadables)
        # make sure we have some files
        assert r.file_handlers

    @pytest.mark.parametrize(
        ("filenames", "expected_datasets"),
        [
            (
                EXAMPLE_FILENAMES,
                [
                    "swath_var_hi",
                    "swath_var_low",
                    "swath_var_low_int",
                    "swath_var",
                    "swath_var_int",
                    "grid_var_NH",
                    "grid_var_SH",
                    "latency_var_NH",
                    "latency_var_SH",
                ],
            ),
            ([MBT_FILENAME], ["swath_var_hi", "swath_var_low", "swath_var_low_int"]),
            ([PRECIP_FILENAME], ["swath_var_hi", "swath_var_low", "swath_var_low_int"]),
            ([OCEAN_FILENAME], ["swath_var_hi", "swath_var_low", "swath_var_low_int"]),
            ([SEAICE_NH_FILENAME], ["grid_var_NH", "latency_var_NH"]),
            ([SEAICE_SH_FILENAME], ["grid_var_SH", "latency_var_SH"]),
            ([SNOW_FILENAME], ["swath_var", "swath_var_int"]),
            ([SOIL_FILENAME], ["swath_var", "swath_var_int"]),
        ],
    )
    def test_available_datasets(self, tmp_path, filenames, expected_datasets):
        """Test that variables are dynamically discovered."""
        from satpy.readers.core.loading import load_reader

        filenames = [tmp_path / filename for filename in filenames]
        for fn in filenames:
            _create_fake_gaasp_file(fn)
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames(filenames)
        r.create_filehandlers(loadables)
        avails = list(r.available_dataset_names)
        for var_name in expected_datasets:
            assert var_name in avails
        assert "not_xy_dim_var" not in expected_datasets

    @staticmethod
    def _check_area(data_id, data_arr):
        from pyresample.geometry import AreaDefinition, SwathDefinition

        area = data_arr.attrs["area"]
        if "grid_var" in data_id["name"] or "latency_var" in data_id["name"]:
            assert isinstance(area, AreaDefinition)
        else:
            assert isinstance(area, SwathDefinition)

    @staticmethod
    def _check_fill(data_id, data_arr):
        if "int" in data_id["name"]:
            assert data_arr.attrs["_FillValue"] == 100
            assert np.issubdtype(data_arr.dtype, np.integer)
        else:
            assert "_FillValue" not in data_arr.attrs
            if data_id["name"].startswith("latency"):
                # timedelta only fits in 64-bit float
                assert data_arr.dtype.type == np.float64
            elif np.issubdtype(data_arr.dtype, np.floating):
                # we started with float32, it should stay that way
                assert data_arr.dtype.type == np.float32

    @staticmethod
    def _check_attrs(data_arr):
        attrs = data_arr.attrs
        assert "scale_factor" not in attrs
        assert "add_offset" not in attrs
        assert attrs["platform_name"] == "GCOM-W1"
        assert attrs["sensor"] == "amsr2"
        assert attrs["start_time"] == dt.datetime(2020, 8, 12, 5, 58, 31)
        assert attrs["end_time"] == dt.datetime(2020, 8, 12, 6, 7, 1)

    @pytest.mark.parametrize(
        ("filenames", "loadable_ids"),
        [
            (
                EXAMPLE_FILENAMES,
                [
                    "swath_var_hi",
                    "swath_var_low",
                    "swath_var_low_int",
                    "swath_var",
                    "swath_var_int",
                    "grid_var_NH",
                    "grid_var_SH",
                    "latency_var_NH",
                    "latency_var_SH",
                ],
            ),
            ([MBT_FILENAME], ["swath_var_hi", "swath_var_low", "swath_var_low_int"]),
            ([PRECIP_FILENAME], ["swath_var_hi", "swath_var_low", "swath_var_low_int"]),
            ([OCEAN_FILENAME], ["swath_var_hi", "swath_var_low", "swath_var_low_int"]),
            ([SEAICE_NH_FILENAME], ["grid_var_NH", "latency_var_NH"]),
            ([SEAICE_SH_FILENAME], ["grid_var_SH", "latency_var_SH"]),
            ([SNOW_FILENAME], ["swath_var", "swath_var_int"]),
            ([SOIL_FILENAME], ["swath_var", "swath_var_int"]),
        ],
    )
    def test_basic_load(self, tmp_path, filenames, loadable_ids):
        """Test that variables are loaded properly."""
        from satpy.readers.core.loading import load_reader

        filenames = [tmp_path / filename for filename in filenames]
        for fn in filenames:
            _create_fake_gaasp_file(fn)
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames(filenames)
        r.create_filehandlers(loadables)
        loaded_data_arrs = r.load(loadable_ids)
        assert loaded_data_arrs
        for data_id, data_arr in loaded_data_arrs.items():
            self._check_area(data_id, data_arr)
            self._check_fill(data_id, data_arr)
            self._check_attrs(data_arr)

    @pytest.mark.parametrize("filter_wspd", [False, True])
    def test_wind_speed_filtering(self, tmp_path, filter_wspd):
        """Test that wind speed can be filtered."""
        from satpy.readers.core.loading import load_reader

        _create_fake_ocean_wind_speed_file(tmp_path / OCEAN_FILENAME)
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([tmp_path / OCEAN_FILENAME])
        r.create_filehandlers(loadables, fh_kwargs={"filter_wind_speed": filter_wspd})
        loaded_data_arrs = r.load(["WSPD"])
        assert loaded_data_arrs
        data_arr = loaded_data_arrs["WSPD"]
        data_id = data_arr.attrs["_satpy_id"]
        self._check_area(data_id, data_arr)
        self._check_fill(data_id, data_arr)
        self._check_attrs(data_arr)
        has_nan = np.isnan(data_arr.data.compute()[-1, -1])
        if filter_wspd:
            assert has_nan
        else:
            assert not has_nan



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for MODIS L3 HDF reader."""

from __future__ import annotations

import dask.array as da
import numpy as np
import pytest
from pyresample import geometry
from pytest_lazy_fixtures import lf as lazy_fixture

from satpy import Scene, available_readers


def _expected_area():
    proj_param = "EPSG:4326"

    return geometry.AreaDefinition("gridded_modis",
                                   "A gridded L3 MODIS area",
                                   "longlat",
                                   proj_param,
                                   7200,
                                   3600,
                                   (-180, -90, 180, 90))


class TestModisL3:
    """Test MODIS L3 reader."""

    def test_available_reader(self):
        """Test that MODIS L3 reader is available."""
        assert "modis_l3" in available_readers()

    @pytest.mark.parametrize(
        ("loadable", "filename"),
        [
            ("Coarse_Resolution_Surface_Reflectance_Band_2", lazy_fixture("modis_l3_nasa_mod09_file")),
            ("BRDF_Albedo_Parameter1_Band2", lazy_fixture("modis_l3_nasa_mod43_file")),
        ]
    )
    def test_scene_available_datasets(self, loadable, filename):
        """Test that datasets are available."""
        scene = Scene(reader="modis_l3", filenames=filename)
        available_datasets = scene.all_dataset_names()
        assert len(available_datasets) > 0
        assert loadable in available_datasets

        from satpy.readers.modis_l3 import ModisL3GriddedHDFFileHandler
        fh = ModisL3GriddedHDFFileHandler(filename[0], {}, {"file_type": "modis_l3_cmg_hdf"})
        configured_datasets = [[None, {"name": "none_ds", "file_type": "modis_l3_cmg_hdf"}],
                               [True, {"name": "true_ds", "file_type": "modis_l3_cmg_hdf"}],
                               [False, {"name": "false_ds", "file_type": "modis_l3_cmg_hdf"}],
                               [None, {"name": "other_ds", "file_type": "modis_l2_random"}]]
        for status, mda in fh.available_datasets(configured_datasets):
            if mda["name"] == "none_ds":
                assert mda["file_type"] == "modis_l3_cmg_hdf"
                assert status is False
            elif mda["name"] == "true_ds":
                assert mda["file_type"] == "modis_l3_cmg_hdf"
                assert status
            elif mda["name"] == "false_ds":
                assert mda["file_type"] == "modis_l3_cmg_hdf"
                assert status is False
            elif mda["name"] == "other_ds":
                assert mda["file_type"] == "modis_l2_random"
                assert status is None
            elif mda["name"] == loadable:
                assert mda["file_type"] == "modis_l3_cmg_hdf"
                assert status

    def test_load_l3_dataset(self, modis_l3_nasa_mod09_file):
        """Load and check an L2 variable."""
        scene = Scene(reader="modis_l3", filenames=modis_l3_nasa_mod09_file)

        ds_name = "Coarse_Resolution_Surface_Reflectance_Band_2"
        scene.load([ds_name])

        data_arr = scene[ds_name]
        assert isinstance(data_arr.data, da.Array)
        data_arr_comp = data_arr.compute()

        # Check types
        assert data_arr_comp.dtype == data_arr.dtype
        assert data_arr_comp.dtype == np.float32

        assert data_arr_comp.shape == (3600, 7200)
        assert data_arr_comp.attrs.get("resolution") == 0.05
        assert data_arr_comp.attrs.get("area") == _expected_area()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for compatible netCDF/Zarr DataArray encodings."""
import datetime

import pytest
import xarray as xr


class TestUpdateEncoding:
    """Test update of dataset encodings."""

    @pytest.fixture
    def fake_ds(self):
        """Create fake data for testing."""
        ds = xr.Dataset({"foo": (("y", "x"), [[1, 2], [3, 4]]),
                         "bar": (("y", "x"), [[3, 4], [5, 6]])},
                        coords={"y": [1, 2],
                                "x": [3, 4],
                                "lon": (("y", "x"), [[7, 8], [9, 10]])})
        return ds

    @pytest.fixture
    def fake_ds_digit(self):
        """Create fake data for testing."""
        ds_digit = xr.Dataset({"CHANNEL_1": (("y", "x"), [[1, 2], [3, 4]]),
                               "CHANNEL_2": (("y", "x"), [[3, 4], [5, 6]])},
                              coords={"y": [1, 2],
                                      "x": [3, 4],
                                      "lon": (("y", "x"), [[7, 8], [9, 10]])})
        return ds_digit

    def test_dataset_name_digit(self, fake_ds_digit):
        """Test data with dataset name staring with a digit."""
        from satpy.cf.encoding import update_encoding

        # Dataset with name staring with digit
        ds_digit = fake_ds_digit
        kwargs = {"encoding": {"1": {"dtype": "float32"},
                               "2": {"dtype": "float32"}},
                  "other": "kwargs"}
        enc, other_kwargs = update_encoding(ds_digit, kwargs, numeric_name_prefix="CHANNEL_")
        expected_dict = {
            "y": {"_FillValue": None},
            "x": {"_FillValue": None},
            "CHANNEL_1": {"dtype": "float32"},
            "CHANNEL_2": {"dtype": "float32"}
        }
        assert enc == expected_dict
        assert other_kwargs == {"other": "kwargs"}

    def test_without_time(self, fake_ds):
        """Test data with no time dimension."""
        from satpy.cf.encoding import update_encoding

        # Without time dimension
        ds = fake_ds.chunk(2)
        kwargs = {"encoding": {"bar": {"chunksizes": (1, 1)}},
                  "other": "kwargs"}
        enc, other_kwargs = update_encoding(ds, kwargs)
        expected_dict = {
            "y": {"_FillValue": None},
            "x": {"_FillValue": None},
            "lon": {"chunksizes": (2, 2)},
            "foo": {"chunksizes": (2, 2)},
            "bar": {"chunksizes": (1, 1)}
        }
        assert enc == expected_dict
        assert other_kwargs == {"other": "kwargs"}

        # Chunksize may not exceed shape
        ds = fake_ds.chunk(8)
        kwargs = {"encoding": {}, "other": "kwargs"}
        enc, other_kwargs = update_encoding(ds, kwargs)
        expected_dict = {
            "y": {"_FillValue": None},
            "x": {"_FillValue": None},
            "lon": {"chunksizes": (2, 2)},
            "foo": {"chunksizes": (2, 2)},
            "bar": {"chunksizes": (2, 2)}
        }
        assert enc == expected_dict

    def test_with_time(self, fake_ds):
        """Test data with a time dimension."""
        from satpy.cf.encoding import update_encoding

        # With time dimension
        ds = fake_ds.chunk(8).expand_dims({"time": [datetime.datetime(2009, 7, 1, 12, 15)]})
        kwargs = {"encoding": {"bar": {"chunksizes": (1, 1, 1)}},
                  "other": "kwargs"}
        enc, other_kwargs = update_encoding(ds, kwargs)
        expected_dict = {
            "y": {"_FillValue": None},
            "x": {"_FillValue": None},
            "lon": {"chunksizes": (2, 2)},
            "foo": {"chunksizes": (1, 2, 2)},
            "bar": {"chunksizes": (1, 1, 1)},
            "time": {"_FillValue": None,
                     "calendar": "proleptic_gregorian",
                     "units": "days since 2009-07-01 12:15:00"},
            "time_bnds": {"_FillValue": None,
                          "calendar": "proleptic_gregorian",
                          "units": "days since 2009-07-01 12:15:00"}
        }
        assert enc == expected_dict
        # User-defined encoding may not be altered
        assert kwargs["encoding"] == {"bar": {"chunksizes": (1, 1, 1)}}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Helpers for reading hdf4-based files."""

import logging
import os

import dask.array as da
import numpy as np
import xarray as xr
from dask.base import tokenize
from pyhdf.SD import SD, SDC, SDS

from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.utils import get_legacy_chunk_size

LOG = logging.getLogger(__name__)

CHUNK_SIZE = get_legacy_chunk_size()

HTYPE_TO_DTYPE = {
    SDC.INT8: np.int8,
    SDC.UCHAR: np.uint8,
    SDC.CHAR: np.int8,
    SDC.INT32: np.int32,
    SDC.INT16: np.int16,
    SDC.UINT8: np.uint8,
    SDC.UINT16: np.uint16,
    SDC.UINT32: np.uint32,
    SDC.FLOAT32: np.float32,
    SDC.FLOAT64: np.float64,
}


def from_sds(var, src_path, **kwargs):
    """Create a dask array from a SD dataset."""
    var_info = var.info()
    var.__dict__["dtype"] = np.dtype(HTYPE_TO_DTYPE[var_info[3]])
    shape = var_info[2]
    var.__dict__["shape"] = shape if isinstance(shape, (tuple, list)) else tuple(shape)

    name = kwargs.pop("name", None)
    if name is None:
        var_name = var_info[0]
        tokenize_args = (os.fspath(src_path), var_name)
        if kwargs:
            tokenize_args += (kwargs,)
        # put variable name in the front for easier dask debugging
        name = var_name + "-" + tokenize(*tokenize_args)
    return da.from_array(var, name=name, **kwargs)


class HDF4FileHandler(BaseFileHandler):
    """Base class for common HDF4 operations."""

    def __init__(self, filename, filename_info, filetype_info):
        """Open file and collect information."""
        super(HDF4FileHandler, self).__init__(filename, filename_info, filetype_info)
        self.file_content = {}
        file_handle = SD(self.filename, SDC.READ)
        self._collect_attrs("", file_handle.attributes())
        for k in file_handle.datasets().keys():
            self.collect_metadata(k, file_handle.select(k))
        del file_handle

    def _collect_attrs(self, name, attrs):
        for key, value in attrs.items():
            value = np.squeeze(value)
            if issubclass(value.dtype.type, (np.bytes_, np.str_)) and not value.shape:
                value = value.item()  # convert to scalar
                if not isinstance(value, str):
                    # python 3 - was scalar numpy array of bytes
                    # otherwise python 2 - scalar numpy array of 'str'
                    value = value.decode()
                self.file_content["{}/attr/{}".format(name, key)] = value
            elif not value.shape:
                # convert to a scalar
                self.file_content["{}/attr/{}".format(name, key)] = value.item()
            else:
                self.file_content["{}/attr/{}".format(name, key)] = value

    def collect_metadata(self, name, obj):
        """Collect all metadata about file content."""
        if isinstance(obj, SDS):
            self.file_content[name] = obj
            info = obj.info()
            self.file_content[name + "/dtype"] = np.dtype(HTYPE_TO_DTYPE.get(info[3]))
            self.file_content[name + "/shape"] = info[2] if isinstance(info[2], (int, float)) else tuple(info[2])

    def _open_xarray_dataset(self, val, chunks=CHUNK_SIZE):
        """Read the band in blocks."""
        dask_arr = from_sds(val, self.filename, chunks=chunks)
        attrs = val.attributes()
        return xr.DataArray(dask_arr, dims=("y", "x"),
                            attrs=attrs)

    def __getitem__(self, key):
        """Get file content as xarray compatible objects."""
        val = self.file_content[key]
        if isinstance(val, SDS):
            # these datasets are closed and inaccessible when the file is closed, need to reopen
            return self._open_xarray_dataset(val)
        return val

    def __contains__(self, item):
        """Check if item is in file content."""
        return item in self.file_content

    def get(self, item, default=None):
        """Get variable as DataArray or return the default."""
        if item in self:
            return self[item]
        else:
            return default



# Copyright (c) 2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.osisaf_l3 module."""

import datetime as dt
import os

import numpy as np
import pytest
import xarray as xr
from pyproj import CRS

from satpy import DataQuery
from satpy.readers.osisaf_l3_nc import OSISAFL3NCFileHandler

stere_ds = xr.DataArray(
    -999,
    attrs={"grid_mapping_name": "polar_stereographic",
           "false_easting": 0.0,
           "false_northing": 0.0,
           "semi_major_axis": 6378273.0,
           "semi_minor_axis": 6356889.44891,
           "straight_vertical_longitude_from_pole": 0.0,
           "latitude_of_projection_origin": -90.0,
           "standard_parallel": -70.0,
           "proj4_string": "+proj=stere +a=6378273 +b=6356889.44891 +lat_0=-90 +lat_ts=-70 +lon_0=0",
           })

stere_ds_noproj = xr.DataArray(
    -999,
    attrs={"grid_mapping_name": "polar_stereographic",
           "false_easting": 0.0,
           "false_northing": 0.0,
           "semi_major_axis": 6378273.0,
           "semi_minor_axis": 6356889.44891,
           "straight_vertical_longitude_from_pole": 0.0,
           "latitude_of_projection_origin": -90.0,
           "standard_parallel": -70.0,
           })

ease_ds = xr.DataArray(
    -999,
    attrs={"grid_mapping_name": "lambert_azimuthal_equal_area",
           "false_easting": 0.0,
           "false_northing": 0.0,
           "semi_major_axis": 6371228.0,
           "longitude_of_projection_origin": 0.0,
           "latitude_of_projection_origin": -90.0,
           "proj4_string": "+proj=laea +a=6371228.0 +lat_0=-90 +lon_0=0",
           })

attrs_ice = {
    "start_date": "2022-12-15 00:00:00",
    "stop_date": "2022-12-16 00:00:00",
    "platform_name": "Multi-sensor analysis",
    "instrument_type": "Multi-sensor analysis"}

attrs_flux = {
    "time_coverage_start": "2023-10-10T00:00:00Z",
    "time_coverage_end": "2023-10-10T23:59:59Z",
    "platform": "NOAA-19, NOAA-20, Metop-B, Metop-C, SNPP",
    "sensor": "AVHRR, VIIRS, AVHRR, AVHRR, VIIRS"}

attrs_geo = {
    "start_time": "20221228T183000Z",
    "stop_time": "20221228T193000Z",
    "platform": "MSG4"}


class OSISAFL3ReaderTests:
    """Test OSI-SAF level 3 netCDF reader ice files."""

    def setup_method(self, tester="ice"):
        """Create a fake dataset."""
        base_data = np.array(([-999, 1215, 1125, 11056, 9500], [200, 1, -999, 4215, 5756]))
        base_data_ssi = np.array(([-999.99, 121.5, 11.25, 110.56, 950.0], [200, 1, -999.99, 42.15, 5.756]))
        base_data_sst = np.array(([-32768, 273.2, 194.2, 220.78, 301.], [-32768, -32768, 273.22, 254.34, 204.21]))
        base_data_ssi_geo = np.array(([-32768, 121.5, 11.25, 110.56, 950.0], [200, 1, -32768, 42.15, 5.756]))
        base_data = np.expand_dims(base_data, axis=0)
        base_data_ssi = np.expand_dims(base_data_ssi, axis=0)
        base_data_sst = np.expand_dims(base_data_sst, axis=0)
        unc_data = np.array(([0, 1, 2, 3, 4], [4, 3, 2, 1, 0]))
        yc_data = np.array(([-10, -5, 0, 5, 10], [-10, -5, 0, 5, 10]))
        xc_data = np.array(([-5, -5, -5, -5, -5], [5, 5, 5, 5, 5]))
        time_data = np.array([1.])
        self.scl = 1.
        self.add = 0.

        lat_data = np.array(([-68, -69, -70, -71, -72], [-68, -69, -70, -71, -72]))
        lon_data = np.array(([-60, -60, -60, -60, -60], [-65, -65, -65, -65, -65]))

        xc = xr.DataArray(xc_data, dims=("yc", "xc"),
                          attrs={"standard_name": "projection_x_coordinate", "units": "km"})
        yc = xr.DataArray(yc_data, dims=("yc", "xc"),
                          attrs={"standard_name": "projection_y_coordinate", "units": "km"})
        time = xr.DataArray(time_data, dims="time",
                            attrs={"standard_name": "projection_y_coordinate", "units": "km"})
        lat = xr.DataArray(lat_data, dims=("yc", "xc"),
                           attrs={"standard_name": "latitude", "units": "degrees_north"})
        lon = xr.DataArray(lon_data, dims=("yc", "xc"),
                           attrs={"standard_name": "longitude", "units": "degrees_east"})
        conc = xr.DataArray(base_data, dims=("time", "yc", "xc"),
                            attrs={"scale_factor": 0.01, "add_offset": 0., "_FillValue": -999, "units": "%",
                                   "valid_min": 0, "valid_max": 10000, "standard_name": "sea_ice_area_fraction"})
        uncert = xr.DataArray(unc_data, dims=("yc", "xc"),
                              attrs={"scale_factor": 0.01, "add_offset": 0., "_FillValue": -999, "valid_min": 0,
                                     "valid_max": 10000, "standard_name": "total_uncertainty"})
        ssi_geo = xr.DataArray(base_data_ssi_geo, dims=("lat", "lon"),
                               attrs={"scale_factor": 0.1, "add_offset": 0., "_FillValue": 32768, "valid_min": 0.,
                                      "valid_max": 1000., "standard_name": "surface_downwelling_shortwave_flux_in_air"})
        ssi = xr.DataArray(base_data_ssi, dims=("time", "yc", "xc"),
                           attrs={"_FillValue": -999.99, "units": "W m-2", "valid_min": 0., "valid_max": 1000.,
                                  "standard_name": "surface_downwelling_shortwave_flux_in_air"})
        sst = xr.DataArray(base_data_sst, dims=("time", "yc", "xc"),
                           attrs={"scale_factor": 0.01, "add_offset": 273.15, "_FillValue": -32768, "units": "K",
                                  "valid_min": -8000., "valid_max": 5000.,
                                  "standard_name": "sea_ice_surface_temperature"})
        data_vars = {"xc": xc, "yc": yc, "time": time, "lat": lat, "lon": lon}

        if tester == "ice":
            data_vars["Lambert_Azimuthal_Grid"] = ease_ds
            data_vars["Polar_Stereographic_Grid"] = stere_ds
            data_vars["ice_conc"] = conc
            data_vars["total_uncertainty"] = uncert
            self.fake_dataset = xr.Dataset(data_vars=data_vars, attrs=attrs_ice)
        elif tester == "sst":
            data_vars["Polar_Stereographic_Grid"] = stere_ds
            data_vars["surface_temperature"] = sst
            self.fake_dataset = xr.Dataset(data_vars=data_vars, attrs=attrs_ice)
        elif tester == "flux_stere":
            data_vars["Polar_Stereographic_Grid"] = stere_ds_noproj
            data_vars["ssi"] = ssi
            self.fake_dataset = xr.Dataset(data_vars=data_vars, attrs=attrs_flux)
        elif tester == "flux_geo":
            data_vars["ssi"] = ssi_geo
            self.fake_dataset = xr.Dataset(data_vars=data_vars, attrs=attrs_geo)

    def test_instantiate_single_netcdf_file(self, tmp_path):
        """Test initialization of file handlers - given a single netCDF file."""
        tmp_filepath = tmp_path / "fake_dataset.nc"
        self.fake_dataset.to_netcdf(os.fspath(tmp_filepath))

        OSISAFL3NCFileHandler(os.fspath(tmp_filepath), self.filename_info, self.filetype_info)

    def test_get_dataset(self, tmp_path):
        """Test retrieval of datasets."""
        tmp_filepath = tmp_path / "fake_dataset.nc"
        self.fake_dataset.to_netcdf(os.fspath(tmp_filepath))

        test = OSISAFL3NCFileHandler(os.fspath(tmp_filepath), self.filename_info, self.filetype_info)

        res = test.get_dataset(DataQuery(name=self.varname), {"standard_name": self.stdname})
        # Check we remove singleton dimension
        assert res.shape[0] == 2
        assert res.shape[1] == 5

        # Test values are correct
        test_ds = self.fake_dataset[self.varname].values.squeeze()
        test_ds = np.where(test_ds == self.fillv, np.nan, test_ds)
        test_ds = np.where(test_ds > self.maxv, np.nan, test_ds)
        test_ds = test_ds / self.scl + self.add
        np.testing.assert_allclose(res.values, test_ds)

        with pytest.raises(KeyError):
            test.get_dataset(DataQuery(name="erroneous dataset"), {"standard_name": "erroneous dataset"})

    def test_get_start_and_end_times(self, tmp_path):
        """Test retrieval of the sensor name from the netCDF file."""
        tmp_filepath = tmp_path / "fake_dataset.nc"
        self.fake_dataset.to_netcdf(os.fspath(tmp_filepath))

        test = OSISAFL3NCFileHandler(os.fspath(tmp_filepath), self.filename_info, self.filetype_info)

        assert test.start_time == self.good_start_time
        assert test.end_time == self.good_stop_time

    def test_get_area_def_bad(self, tmp_path):
        """Test getting the area definition for the polar stereographic grid."""
        filename_info = {"grid": "turnips"}
        tmp_filepath = tmp_path / "fake_dataset.nc"
        self.fake_dataset.to_netcdf(os.fspath(tmp_filepath))

        test = OSISAFL3NCFileHandler(os.fspath(tmp_filepath), filename_info, self.filetype_info)
        with pytest.raises(ValueError, match="Unknown grid type: turnips"):
            test.get_area_def(None)


class TestOSISAFL3ReaderICE(OSISAFL3ReaderTests):
    """Test OSI-SAF level 3 netCDF reader ice files."""

    def setup_method(self):
        """Set up the tests."""
        super().setup_method(tester="ice")
        self.filename_info = {"grid": "ease"}
        self.filetype_info = {"file_type": "osi_sea_ice_conc"}
        self.good_start_time = dt.datetime(2022, 12, 15, 0, 0, 0)
        self.good_stop_time = dt.datetime(2022, 12, 16, 0, 0, 0)
        self.varname = "ice_conc"
        self.stdname = "sea_ice_area_fraction"
        self.fillv = -999
        self.maxv = 10000
        self.scl = 100

    def test_get_area_def_stere(self, tmp_path):
        """Test getting the area definition for the polar stereographic grid."""
        self.filename_info = {"grid": "stere"}
        tmp_filepath = tmp_path / "fake_dataset.nc"
        self.fake_dataset.to_netcdf(os.fspath(tmp_filepath))

        test = OSISAFL3NCFileHandler(os.fspath(tmp_filepath), self.filename_info, self.filetype_info)

        area_def = test.get_area_def(None)
        assert area_def.description == "osisaf_polar_stereographic"

        expected_crs = CRS(dict(a=6378273.0, lat_0=-90, lat_ts=-70, lon_0=0, proj="stere", rf=298.27940986765))
        assert area_def.crs == expected_crs

        assert area_def.width == 5
        assert area_def.height == 2
        np.testing.assert_allclose(area_def.area_extent,
                                   (-2185821.7955, 1019265.4426, -1702157.4538, 982741.0642))

    def test_get_area_def_ease(self, tmp_path):
        """Test getting the area definition for the EASE grid."""
        tmp_filepath = tmp_path / "fake_dataset.nc"
        self.fake_dataset.to_netcdf(os.fspath(tmp_filepath))

        test = OSISAFL3NCFileHandler(os.fspath(tmp_filepath), {"grid": "ease"}, self.filetype_info)

        area_def = test.get_area_def(None)
        assert area_def.description == "osisaf_lambert_azimuthal_equal_area"

        expected_crs = CRS(dict(R=6371228, lat_0=-90, lon_0=0, proj="laea"))
        assert area_def.crs == expected_crs

        assert area_def.width == 5
        assert area_def.height == 2
        np.testing.assert_allclose(area_def.area_extent,
                                   (-2203574.302335, 1027543.572492, -1726299.781982, 996679.643829))


class TestOSISAFL3ReaderFluxStere(OSISAFL3ReaderTests):
    """Test OSI-SAF level 3 netCDF reader flux files on stereographic grid."""

    def setup_method(self):
        """Set up the tests."""
        super().setup_method(tester="flux_stere")
        self.filename_info = {"grid": "polstere"}
        self.filetype_info = {"file_type": "osi_radflux_stere"}
        self.good_start_time = dt.datetime(2023, 10, 10, 0, 0, 0)
        self.good_stop_time = dt.datetime(2023, 10, 10, 23, 59, 59)
        self.varname = "ssi"
        self.stdname = "surface_downwelling_shortwave_flux_in_air"
        self.fillv = -999.99
        self.maxv = 1000
        self.scl = 1

    def test_get_area_def_stere(self, tmp_path):
        """Test getting the area definition for the polar stereographic grid."""
        tmp_filepath = tmp_path / "fake_dataset.nc"
        self.fake_dataset.to_netcdf(os.fspath(tmp_filepath))

        test = OSISAFL3NCFileHandler(os.fspath(tmp_filepath), self.filename_info, self.filetype_info)

        area_def = test.get_area_def(None)
        assert area_def.description == "osisaf_polar_stereographic"

        expected_crs = CRS(dict(a=6378273.0, lat_0=-90, lat_ts=-70, lon_0=0, proj="stere", rf=298.27940986765))
        assert area_def.crs == expected_crs

        assert area_def.width == 5
        assert area_def.height == 2
        np.testing.assert_allclose(area_def.area_extent,
                                   (-2185821.7955, 1019265.4426, -1702157.4538, 982741.0642))


class TestOSISAFL3ReaderFluxGeo(OSISAFL3ReaderTests):
    """Test OSI-SAF level 3 netCDF reader flux files on lat/lon grid (GEO sensors)."""

    def setup_method(self):
        """Set up the tests."""
        super().setup_method(tester="flux_geo")
        self.filename_info = {}
        self.filetype_info = {"file_type": "osi_radflux_grid"}
        self.good_start_time = dt.datetime(2022, 12, 28, 18, 30, 0)
        self.good_stop_time = dt.datetime(2022, 12, 28, 19, 30, 0)
        self.varname = "ssi"
        self.stdname = "surface_downwelling_shortwave_flux_in_air"
        self.fillv = -32768
        self.maxv = 1000
        self.scl = 10

    def test_get_area_def_grid(self, tmp_path):
        """Test getting the area definition for the lat/lon grid."""
        tmp_filepath = tmp_path / "fake_dataset.nc"
        self.filename_info = {}
        self.filetype_info = {"file_type": "osi_radflux_grid"}
        self.fake_dataset.to_netcdf(os.fspath(tmp_filepath))

        test = OSISAFL3NCFileHandler(os.fspath(tmp_filepath), self.filename_info, self.filetype_info)

        area_def = test.get_area_def(None)
        assert area_def.description == "osisaf_geographic_area"

        expected_crs = CRS(dict(datum="WGS84", proj="longlat"))
        assert area_def.crs == expected_crs

        assert area_def.width == 5
        assert area_def.height == 2
        np.testing.assert_allclose(area_def.area_extent,
                                   (-65, -68, -60, -72))


class TestOSISAFL3ReaderSST(OSISAFL3ReaderTests):
    """Test OSI-SAF level 3 netCDF reader surface temperature files."""

    def setup_method(self):
        """Set up the tests."""
        super().setup_method(tester="sst")
        self.filename_info = {}
        self.filetype_info = {"file_type": "osi_sst"}
        self.good_start_time = dt.datetime(2022, 12, 15, 0, 0, 0)
        self.good_stop_time = dt.datetime(2022, 12, 16, 0, 0, 0)
        self.varname = "surface_temperature"
        self.stdname = "sea_ice_surface_temperature"
        self.fillv = -32768
        self.maxv = 1000
        self.scl = 100
        self.add = 273.15

    def test_get_area_def_stere(self, tmp_path):
        """Test getting the area definition for the polar stereographic grid."""
        tmp_filepath = tmp_path / "fake_dataset.nc"
        self.fake_dataset.to_netcdf(os.fspath(tmp_filepath))

        test = OSISAFL3NCFileHandler(os.fspath(tmp_filepath), self.filename_info, self.filetype_info)

        area_def = test.get_area_def(None)
        assert area_def.description == "osisaf_polar_stereographic"

        expected_crs = CRS(dict(a=6378273.0, lat_0=-90, lat_ts=-70, lon_0=0, proj="stere", rf=298.27940986765))

        assert area_def.crs == expected_crs

        assert area_def.width == 5
        assert area_def.height == 2
        np.testing.assert_allclose(area_def.area_extent,
                                   (-2185821.7955, 1019265.4426, -1702157.4538, 982741.0642))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the geotiff writer."""

import datetime as dt
from unittest import mock

import dask.array as da
import numpy as np
import pytest
import xarray as xr

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - tmp_path


def _get_test_datasets_2d():
    """Create a single 2D test dataset."""
    from pyresample.geometry import AreaDefinition

    adef = AreaDefinition(
        "test", "test", "test", "EPSG:4326",
        100, 200, (-180., -90., 180., 90.),
    )
    ds1 = xr.DataArray(
        da.arange(100 * 200, dtype=np.float32).reshape((100, 200)).rechunk(50),
        dims=("y", "x"),
        attrs={"name": "test",
               "start_time": dt.datetime.now(dt.timezone.utc),
               "units": "K",
               "area": adef}
    )
    return [ds1]


def _get_test_datasets_2d_nonlinear_enhancement():
    data_arrays = _get_test_datasets_2d()
    enh_history = [
        {"gamma": 2.0},
    ]
    for data_arr in data_arrays:
        data_arr.attrs["enhancement_history"] = enh_history
    return data_arrays


def _get_test_datasets_3d():
    """Create a single 3D test dataset."""
    from pyresample.geometry import AreaDefinition

    adef = AreaDefinition(
        "test", "test", "test", "EPSG:4326",
        100, 200, (-180., -90., 180., 90.),
    )
    ds1 = xr.DataArray(
        da.arange(3 * 100 * 200, dtype=np.float32).reshape((3, 100, 200)).rechunk(50),
        dims=("bands", "y", "x"),
        coords={"bands": ["R", "G", "B"]},
        attrs={"name": "test",
               "start_time": dt.datetime.now(dt.timezone.utc),
               "area": adef}
    )
    return [ds1]


class TestGeoTIFFWriter:
    """Test the GeoTIFF Writer class."""

    def test_init(self):
        """Test creating the writer with no arguments."""
        from satpy.writers.geotiff import GeoTIFFWriter
        GeoTIFFWriter()

    @pytest.mark.parametrize(
        "input_func",
        [
            _get_test_datasets_2d,
            _get_test_datasets_3d
        ]
    )
    def test_simple_write(self, input_func, tmp_path):
        """Test basic writer operation."""
        from satpy.writers.geotiff import GeoTIFFWriter
        datasets = input_func()
        w = GeoTIFFWriter(base_dir=tmp_path)
        w.save_datasets(datasets)

    def test_simple_delayed_write(self, tmp_path):
        """Test writing can be delayed."""
        from satpy.writers.geotiff import GeoTIFFWriter
        datasets = _get_test_datasets_2d()
        w = GeoTIFFWriter(base_dir=tmp_path)
        # when we switch to rio_save on XRImage then this will be sources
        # and targets
        res = w.save_datasets(datasets, compute=False)
        # this will fail if rasterio isn't installed
        assert isinstance(res, tuple)
        # two lists, sources and destinations
        assert len(res) == 2
        assert isinstance(res[0], list)
        assert isinstance(res[1], list)
        assert isinstance(res[0][0], da.Array)
        da.store(res[0], res[1])
        for target in res[1]:
            if hasattr(target, "close"):
                target.close()

    def test_colormap_write(self, tmp_path):
        """Test writing an image with a colormap."""
        from trollimage.colormap import spectral
        from trollimage.xrimage import XRImage

        from satpy.writers.geotiff import GeoTIFFWriter
        datasets = _get_test_datasets_2d()
        w = GeoTIFFWriter(base_dir=tmp_path)
        # we'd have to customize enhancements to test this through
        # save_datasets. We'll use `save_image` as a workaround.
        img = XRImage(datasets[0])
        img.palettize(spectral)
        w.save_image(img, keep_palette=True)

    def test_float_write(self, tmp_path):
        """Test that geotiffs can be written as floats.

        NOTE: Does not actually check that the output is floats.

        """
        from satpy.writers.geotiff import GeoTIFFWriter
        datasets = _get_test_datasets_2d()
        w = GeoTIFFWriter(base_dir=tmp_path,
                          enhance=False,
                          dtype=np.float32)
        w.save_datasets(datasets)

    def test_dtype_for_enhance_false(self, tmp_path):
        """Test that dtype of dataset is used if parameters enhance=False and dtype=None."""
        from satpy.writers.geotiff import GeoTIFFWriter
        datasets = _get_test_datasets_2d()
        w = GeoTIFFWriter(base_dir=tmp_path, enhance=False)
        with mock.patch("trollimage.xrimage.XRImage.save") as save_method:
            # compute is False in `save_datasets` so we need to return something dask-like
            save_method.return_value = da.zeros((1, 1))
            w.save_datasets(datasets, compute=False)
            assert save_method.call_args[1]["dtype"] == np.float32

    def test_dtype_for_enhance_false_and_given_dtype(self, tmp_path):
        """Test that dtype of dataset is used if enhance=False and dtype=uint8."""
        from satpy.writers.geotiff import GeoTIFFWriter
        datasets = _get_test_datasets_2d()
        w = GeoTIFFWriter(base_dir=tmp_path, enhance=False, dtype=np.uint8)
        with mock.patch("trollimage.xrimage.XRImage.save") as save_method:
            # compute is False in `save_datasets` so we need to return something dask-like
            save_method.return_value = da.zeros((1, 1))
            w.save_datasets(datasets, compute=False)
            assert save_method.call_args[1]["dtype"] == np.uint8

    def test_fill_value_from_config(self, tmp_path):
        """Test fill_value coming from the writer config."""
        from satpy.writers.geotiff import GeoTIFFWriter
        datasets = _get_test_datasets_2d()
        w = GeoTIFFWriter(base_dir=tmp_path)
        w.info["fill_value"] = 128
        with mock.patch("trollimage.xrimage.XRImage.save") as save_method:
            # compute is False in `save_datasets` so we need to return something dask-like
            save_method.return_value = da.zeros((1, 1))
            w.save_datasets(datasets, compute=False)
            assert save_method.call_args[1]["fill_value"] == 128

    def test_tags(self, tmp_path):
        """Test tags being added."""
        from satpy.writers.geotiff import GeoTIFFWriter
        datasets = _get_test_datasets_2d()
        w = GeoTIFFWriter(tags={"test1": 1}, base_dir=tmp_path)
        w.info["fill_value"] = 128
        with mock.patch("trollimage.xrimage.XRImage.save") as save_method:
            # compute is False in `save_datasets` so we need to return something dask-like
            save_method.return_value = da.zeros((1, 1))
            w.save_datasets(datasets, tags={"test2": 2}, compute=False)
            called_tags = save_method.call_args[1]["tags"]
            assert called_tags == {"test1": 1, "test2": 2}

    @pytest.mark.parametrize(
        "input_func",
        [
            _get_test_datasets_2d,
            _get_test_datasets_3d,
            _get_test_datasets_2d_nonlinear_enhancement,
        ]
    )
    @pytest.mark.parametrize(
        "save_kwargs",
        [
            {"include_scale_offset": True},
            {"scale_offset_tags": ("scale", "offset")},
        ]
    )
    def test_scale_offset(self, input_func, save_kwargs, tmp_path):
        """Test tags being added."""
        from satpy.writers.geotiff import GeoTIFFWriter
        datasets = input_func()
        w = GeoTIFFWriter(tags={"test1": 1}, base_dir=tmp_path)
        w.info["fill_value"] = 128
        with mock.patch("trollimage.xrimage.XRImage.save") as save_method:
            # compute is False in `save_datasets` so we need to return something dask-like
            save_method.return_value = da.zeros((1, 1))
            w.save_datasets(datasets, tags={"test2": 2}, compute=False, **save_kwargs)
        kwarg_name = "include_scale_offset_tags" if "include_scale_offset" in save_kwargs else "scale_offset_tags"
        kwarg_value = save_method.call_args[1].get(kwarg_name)
        assert kwarg_value is not None

    def test_tiled_value_from_config(self, tmp_path):
        """Test tiled value coming from the writer config."""
        from satpy.writers.geotiff import GeoTIFFWriter
        datasets = _get_test_datasets_2d()
        w = GeoTIFFWriter(base_dir=tmp_path)
        with mock.patch("trollimage.xrimage.XRImage.save") as save_method:
            # compute is False in `save_datasets` so we need to return something dask-like
            save_method.return_value = da.zeros((1, 1))
            w.save_datasets(datasets, compute=False)
            assert save_method.call_args[1]["tiled"]

    def test_float_write_with_unit_conversion(self, tmp_path):
        """Test that geotiffs can be written as floats and convert units."""
        from satpy.writers.geotiff import GeoTIFFWriter
        dataset = _get_test_datasets_2d()[0]
        dtype = np.float32
        w = GeoTIFFWriter(base_dir=tmp_path,
                          enhance=False,
                          dtype=dtype)
        filename = tmp_path / "data_in_C.tif"
        w.save_dataset(dataset, filename=filename, units="degC")
        ds = xr.open_dataset(filename, engine="rasterio")
        assert ds["band_data"].dtype == dtype
        exp = np.arange(100 * 200, dtype=np.float32).reshape((1, 100, 200)) - 273.15
        np.testing.assert_allclose(ds["band_data"], exp)



#!/usr/bin/env python
# Copyright (c) 2015-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Base resampling functionality."""
import hashlib
import json
import warnings
from contextlib import suppress
from importlib import import_module
from logging import getLogger
from weakref import WeakValueDictionary

import numpy as np

from satpy.utils import get_legacy_chunk_size

LOG = getLogger(__name__)

CHUNK_SIZE = get_legacy_chunk_size()

resamplers_cache: "WeakValueDictionary[tuple, object]" = WeakValueDictionary()


def _hash_dict(the_dict, the_hash=None):
    """Calculate a hash for a dictionary."""
    if the_hash is None:
        the_hash = hashlib.sha1()  # nosec
    the_hash.update(json.dumps(the_dict, sort_keys=True).encode("utf-8"))
    return the_hash


def _update_resampled_coords(old_data, new_data, new_area):
    """Add coordinate information to newly resampled DataArray.

    Args:
        old_data (xarray.DataArray): Old data before resampling.
        new_data (xarray.DataArray): New data after resampling.
        new_area (pyresample.geometry.BaseDefinition): Area definition
            for the newly resampled data.

    """
    from satpy.coords import add_crs_xy_coords

    # copy over other non-x/y coordinates
    # this *MUST* happen before we set 'crs' below otherwise any 'crs'
    # coordinate in the coordinate variables we are copying will overwrite the
    # 'crs' coordinate we just assigned to the data
    ignore_coords = ("y", "x", "crs")
    new_coords = {}
    for cname, cval in old_data.coords.items():
        # we don't want coordinates that depended on the old x/y dimensions
        has_ignored_dims = any(dim in cval.dims for dim in ignore_coords)
        if cname in ignore_coords or has_ignored_dims:
            continue
        new_coords[cname] = cval
    new_data = new_data.assign_coords(**new_coords)

    # add crs, x, and y coordinates
    new_data = add_crs_xy_coords(new_data, new_area)
    return new_data


# TODO: move these to pyresample.resampler
RESAMPLER_MODULES = [
    "satpy.resample.native",
    "satpy.resample.kdtree",
    "satpy.resample.bucket",
    "satpy.resample.ewa",
]

def _get_resampler_classes_from_module(import_path):
    with suppress(ImportError):
        mod = import_module(import_path)
        return mod.get_resampler_classes()
    return {}


def get_all_resampler_classes():
    """Get all available resampler classes."""
    resamplers = {}
    # Collect all available resampler classes
    for import_path in RESAMPLER_MODULES:
        res = _get_resampler_classes_from_module(import_path)
        resamplers.update(res)

    # Add gradient search, which is infact a factory function
    # TODO: add `get_resampler_classes()` function to pyresample.gradient
    with suppress(ImportError):
        from pyresample.gradient import create_gradient_search_resampler
        resamplers["gradient_search"] = create_gradient_search_resampler

    return resamplers


# TODO: move this to pyresample
def prepare_resampler(source_area, destination_area, resampler=None, **resample_kwargs):
    """Instantiate and return a resampler."""
    from pyresample.resampler import BaseResampler as PRBaseResampler

    if resampler is None:
        LOG.info("Using default KDTree resampler")
        resampler = "kd_tree"

    if isinstance(resampler, PRBaseResampler):
        raise ValueError("Trying to create a resampler when one already "
                         "exists.")
    if isinstance(resampler, str):
        resampler_class = get_all_resampler_classes().get(resampler, None)
        _check_resampler_class(resampler_class, resampler)
    else:
        resampler_class = resampler

    key = (resampler_class,
           source_area, destination_area,
           _hash_dict(resample_kwargs).hexdigest())
    try:
        resampler_instance = resamplers_cache[key]
    except KeyError:
        resampler_instance = resampler_class(source_area, destination_area)
        resamplers_cache[key] = resampler_instance
    return key, resampler_instance


def _check_resampler_class(resampler_class, resampler):
    if resampler_class is not None:
        return
    if resampler == "gradient_search":
        warnings.warn(
            "Gradient search resampler not available. Maybe missing `shapely`?",
            stacklevel=2
        )
    raise KeyError("Resampler '%s' not available" % resampler)


# TODO: move this to pyresample
def resample(source_area, data, destination_area,
             resampler=None, **kwargs):
    """Do the resampling."""
    from pyresample.resampler import BaseResampler as PRBaseResampler

    if not isinstance(resampler, PRBaseResampler):
        # we don't use the first argument (cache key)
        _, resampler_instance = prepare_resampler(source_area,
                                                  destination_area,
                                                  resampler)
    else:
        resampler_instance = resampler

    if isinstance(data, list):
        res = [resampler_instance.resample(ds, **kwargs) for ds in data]
    else:
        res = resampler_instance.resample(data, **kwargs)

    return res


def _get_fill_value(dataset):
    """Get the fill value of the *dataset*, defaulting to np.nan."""
    if np.issubdtype(dataset.dtype, np.integer):
        return dataset.attrs.get("_FillValue", np.nan)
    return np.nan


def resample_dataset(dataset, destination_area, **kwargs):
    """Resample *dataset* and return the resampled version.

    Args:
        dataset (xarray.DataArray): Data to be resampled.
        destination_area: The destination onto which to project the data,
          either a full blown area definition or a string corresponding to
          the name of the area as defined in the area file.
        **kwargs: The extra parameters to pass to the resampler objects.

    Returns:
        A resampled DataArray with updated ``.attrs["area"]`` field. The dtype
        of the array is preserved.

    """
    # call the projection stuff here
    try:
        source_area = dataset.attrs["area"]
    except KeyError:
        LOG.info("Cannot reproject dataset %s, missing area info",
                 dataset.attrs["name"])

        return dataset

    fill_value = kwargs.pop("fill_value", _get_fill_value(dataset))
    new_data = resample(source_area, dataset, destination_area, fill_value=fill_value, **kwargs)
    new_attrs = new_data.attrs
    new_data.attrs = dataset.attrs.copy()
    new_data.attrs.update(new_attrs)
    new_data.attrs.update(area=destination_area)

    return new_data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""The abi_l1b reader tests package."""

from __future__ import annotations

import datetime as dt
from pathlib import Path
from typing import Any, Callable
from unittest import mock

import dask
import dask.array as da
import numpy as np
import numpy.typing as npt
import pytest
import xarray as xr
from pytest_lazy_fixtures import lf as lazy_fixture

from satpy import DataQuery
from satpy.readers.abi_l1b import NC_ABI_L1B
from satpy.readers.core.yaml_reader import FileYAMLReader
from satpy.utils import ignore_pyproj_proj_warnings

RAD_SHAPE = {
    500: (3000, 5000),  # conus - 500m
}
RAD_SHAPE[1000] = (RAD_SHAPE[500][0] // 2, RAD_SHAPE[500][1] // 2)
RAD_SHAPE[2000] = (RAD_SHAPE[500][0] // 4, RAD_SHAPE[500][1] // 4)


def _create_fake_rad_dataarray(
    rad: xr.DataArray | None = None,
    resolution: int = 2000,
) -> xr.DataArray:
    x_image = xr.DataArray(0.0)
    y_image = xr.DataArray(0.0)
    time = xr.DataArray(0.0)
    shape = RAD_SHAPE[resolution]
    if rad is None:
        rad_data = (np.arange(shape[0] * shape[1]).reshape(shape) + 1.0) * 50.0
        rad_data = (rad_data + 1.0) / 0.5
        rad_data = rad_data.astype(np.int16)
        rad = xr.DataArray(
            da.from_array(rad_data, chunks=226),
            dims=("y", "x"),
            attrs={
                "scale_factor": 0.5,
                "add_offset": -1.0,
                "_FillValue": 1002,
                "units": "W m-2 um-1 sr-1",
                "valid_range": (0, 4095),
            },
        )
    rad.coords["t"] = time
    rad.coords["x_image"] = x_image
    rad.coords["y_image"] = y_image
    return rad


def _create_fake_rad_dataset(rad: xr.DataArray, resolution: int) -> xr.Dataset:
    rad = _create_fake_rad_dataarray(rad=rad, resolution=resolution)

    x__ = xr.DataArray(
        range(rad.shape[1]),
        attrs={"scale_factor": 2.0, "add_offset": -1.0},
        dims=("x",),
    )
    y__ = xr.DataArray(
        range(rad.shape[0]),
        attrs={"scale_factor": -2.0, "add_offset": 1.0},
        dims=("y",),
    )
    proj = xr.DataArray(
        np.int64(0),
        attrs={
            "semi_major_axis": 1.0,
            "semi_minor_axis": 1.0,
            "perspective_point_height": 1.0,
            "longitude_of_projection_origin": -90.0,
            "latitude_of_projection_origin": 0.0,
            "sweep_angle_axis": "x",
        },
    )

    fake_dataset = xr.Dataset(
        data_vars={
            "Rad": rad,
            "band_id": np.array(8),
            # 'x': x__,
            # 'y': y__,
            "x_image": xr.DataArray(0.0),
            "y_image": xr.DataArray(0.0),
            "goes_imager_projection": proj,
            "yaw_flip_flag": np.array([1]),
            "planck_fk1": np.array(13432.1),
            "planck_fk2": np.array(1497.61),
            "planck_bc1": np.array(0.09102),
            "planck_bc2": np.array(0.99971),
            "esun": np.array(2017),
            "nominal_satellite_subpoint_lat": np.array(0.0),
            "nominal_satellite_subpoint_lon": np.array(-89.5),
            "nominal_satellite_height": np.array(35786.02),
            "earth_sun_distance_anomaly_in_AU": np.array(0.99),
        },
        coords={
            "t": rad.coords["t"],
            "x": x__,
            "y": y__,
        },
        attrs={
            "time_coverage_start": "2017-09-20T17:30:40.8Z",
            "time_coverage_end": "2017-09-20T17:41:17.5Z",
        },
    )
    return fake_dataset


def generate_l1b_filename(chan_name: str) -> str:
    """Generate a l1b filename."""
    return f"OR_ABI-L1b-RadC-M4{chan_name}_G16_s20161811540362_e20161811545170_c20161811545230_suffix.nc"


@pytest.fixture
def c01_refl(tmp_path) -> xr.DataArray:
    """Load c01 reflectances."""
    with _apply_dask_chunk_size():
        reader = _create_reader_for_data(tmp_path, "C01", None, 1000)
        return reader.load(["C01"])["C01"]


@pytest.fixture
def c01_rad(tmp_path) -> xr.DataArray:
    """Load c01 radiances."""
    with _apply_dask_chunk_size():
        reader = _create_reader_for_data(tmp_path, "C01", None, 1000)
        return reader.load([DataQuery(name="C01", calibration="radiance")])["C01"]


@pytest.fixture
def c01_rad_h5netcdf(tmp_path) -> xr.DataArray:
    """Load c01 radiances through h5netcdf."""
    shape = RAD_SHAPE[1000]
    rad_data = (np.arange(shape[0] * shape[1]).reshape(shape) + 1.0) * 50.0
    rad_data = (rad_data + 1.0) / 0.5
    rad_data = rad_data.astype(np.int16)
    rad = xr.DataArray(
        da.from_array(rad_data, chunks=226),
        dims=("y", "x"),
        attrs={
            "scale_factor": 0.5,
            "add_offset": -1.0,
            "_FillValue": np.array([1002]),
            "units": "W m-2 um-1 sr-1",
            "valid_range": (0, 4095),
        },
    )
    with _apply_dask_chunk_size():
        reader = _create_reader_for_data(tmp_path, "C01", rad, 1000)
        return reader.load([DataQuery(name="C01", calibration="radiance")])["C01"]


@pytest.fixture
def c01_counts(tmp_path) -> xr.DataArray:
    """Load c01 counts."""
    with _apply_dask_chunk_size():
        reader = _create_reader_for_data(tmp_path, "C01", None, 1000)
        return reader.load([DataQuery(name="C01", calibration="counts")])["C01"]


@pytest.fixture
def c07_bt_creator(tmp_path) -> Callable:
    """Create a loader for c07 brightness temperatures."""
    def _load_data_array(
        clip_negative_radiances: bool = False,
    ):
        rad = _fake_c07_data()
        with _apply_dask_chunk_size():
            reader = _create_reader_for_data(
                tmp_path,
                "C07",
                rad,
                2000,
                {"clip_negative_radiances": clip_negative_radiances},
            )
            return reader.load(["C07"])["C07"]

    return _load_data_array


def _fake_c07_data() -> xr.DataArray:
    shape = RAD_SHAPE[2000]
    values = np.arange(shape[0] * shape[1])
    rad_data = (values.reshape(shape) + 1.0) * 50.0
    rad_data[0, 0] = -0.0001  # introduce below minimum expected radiance
    rad_data = (rad_data + 1.3) / 0.5
    data = rad_data.astype(np.int16)
    rad = xr.DataArray(
        da.from_array(data, chunks=226),
        dims=("y", "x"),
        attrs={
            "scale_factor": 0.5,
            "add_offset": -1.3,
            "_FillValue": np.int16(
                np.floor(((9 + 1) * 50.0 + 1.3) / 0.5)
            ),  # last rad_data value
        },
    )
    return rad


def _create_reader_for_data(
        tmp_path: Path,
        channel_name: str,
        rad: xr.DataArray | None,
        resolution: int,
        reader_kwargs: dict[str, Any] | None = None,
) -> FileYAMLReader:
    filename = generate_l1b_filename(channel_name)
    data_path = tmp_path / filename
    dataset = _create_fake_rad_dataset(rad=rad, resolution=resolution)
    dataset.to_netcdf(
        data_path,
        encoding={
            "Rad": {"chunksizes": (226, 226)},
        },
    )
    from satpy.readers.core.loading import load_readers
    return load_readers([str(data_path)], "abi_l1b", reader_kwargs=reader_kwargs)["abi_l1b"]


def _apply_dask_chunk_size():
    # 226 on-disk chunk size
    # 8 on-disk chunks for 500 meter data
    # Square (**2) for 2D size
    # 4 bytes for 32-bit floats
    return dask.config.set({"array.chunk-size": ((226 * 8) ** 2) * 4})


def _get_and_check_array(data_arr: xr.DataArray, exp_dtype: npt.DTypeLike) -> npt.NDArray:
    data_np = data_arr.data.compute()
    assert isinstance(data_arr, xr.DataArray)
    assert isinstance(data_arr.data, da.Array)
    assert isinstance(data_np, np.ndarray)
    res = 1000 if RAD_SHAPE[1000][0] == data_np.shape[0] else 2000
    assert data_arr.chunks[0][0] == 226 * (8 / (res / 500))
    assert data_arr.chunks[1][0] == 226 * (8 / (res / 500))

    assert data_np.dtype == data_arr.dtype
    assert data_np.dtype == exp_dtype
    return data_np


def _check_area(data_arr: xr.DataArray) -> None:
    from pyresample.geometry import AreaDefinition

    area_def = data_arr.attrs["area"]
    assert isinstance(area_def, AreaDefinition)

    with ignore_pyproj_proj_warnings():
        proj_dict = area_def.crs.to_dict()
        exp_dict = {
            "h": 1.0,
            "lon_0": -90.0,
            "proj": "geos",
            "sweep": "x",
            "units": "m",
        }
        if "R" in proj_dict:
            assert proj_dict["R"] == 1
        else:
            assert proj_dict["a"] == 1
            assert proj_dict["b"] == 1
        for proj_key, proj_val in exp_dict.items():
            assert proj_dict[proj_key] == proj_val

    assert area_def.shape == data_arr.shape
    if area_def.shape[0] == RAD_SHAPE[1000][0]:
        exp_extent = (-2.0, -2998.0, 4998.0, 2.0)
    else:
        exp_extent = (-2.0, -1498.0, 2498.0, 2.0)
    assert area_def.area_extent == exp_extent


def _check_dims_and_coords(data_arr: xr.DataArray) -> None:
    assert "y" in data_arr.dims
    assert "x" in data_arr.dims

    # we remove any time dimension information
    assert "t" not in data_arr.coords
    assert "t" not in data_arr.dims
    assert "time" not in data_arr.coords
    assert "time" not in data_arr.dims


@pytest.mark.parametrize(
    ("channel", "suffix"),
    [
        ("C{:02d}".format(num), suffix)
        for num in range(1, 17)
        for suffix in ("", "_test_suffix")
    ],
)
def test_file_patterns_match(channel, suffix):
    """Test that the configured file patterns work."""
    from satpy.readers.core.config import configs_for_reader
    from satpy.readers.core.loading import load_reader

    reader_configs = list(configs_for_reader("abi_l1b"))[0]
    reader = load_reader(reader_configs)
    fn1 = (
        "OR_ABI-L1b-RadM1-M3{}_G16_s20182541300210_e20182541300267"
        "_c20182541300308{}.nc"
    ).format(channel, suffix)
    loadables = reader.select_files_from_pathnames([fn1])
    assert len(loadables) == 1
    if not suffix and channel in ["C01", "C02", "C03", "C05"]:
        fn2 = (
            "OR_ABI-L1b-RadM1-M3{}_G16_s20182541300210_e20182541300267"
            "_c20182541300308-000000_0.nc"
        ).format(channel)
        loadables = reader.select_files_from_pathnames([fn2])
        assert len(loadables) == 1


@pytest.mark.parametrize(
    "c01_data_arr", [lazy_fixture("c01_rad"), lazy_fixture("c01_rad_h5netcdf")]
)
class Test_NC_ABI_L1B:
    """Test the NC_ABI_L1B reader."""

    def test_get_dataset(self, c01_data_arr):
        """Test the get_dataset method."""
        exp = {
            "calibration": "radiance",
            "instrument_ID": None,
            "modifiers": (),
            "name": "C01",
            "observation_type": "Rad",
            "orbital_parameters": {
                "projection_altitude": 1.0,
                "projection_latitude": 0.0,
                "projection_longitude": -90.0,
                "satellite_nominal_altitude": 35786020.0,
                "satellite_nominal_latitude": 0.0,
                "satellite_nominal_longitude": -89.5,
                "yaw_flip": True,
            },
            "orbital_slot": None,
            "platform_name": "GOES-16",
            "platform_shortname": "G16",
            "production_site": None,
            "reader": "abi_l1b",
            "resolution": 1000,
            "scan_mode": "M4",
            "scene_abbr": "C",
            "scene_id": None,
            "sensor": "abi",
            "timeline_ID": None,
            "suffix": "suffix",
            "units": "W m-2 um-1 sr-1",
            "start_time": dt.datetime(2017, 9, 20, 17, 30, 40, 800000),
            "end_time": dt.datetime(2017, 9, 20, 17, 41, 17, 500000),
        }

        res = c01_data_arr
        _get_and_check_array(res, np.float32)
        _check_area(res)
        _check_dims_and_coords(res)
        for exp_key, exp_val in exp.items():
            assert res.attrs[exp_key] == exp_val


@pytest.mark.parametrize("clip_negative_radiances", [False, True])
def test_ir_calibrate(c07_bt_creator, clip_negative_radiances):
    """Test IR calibration."""
    res = c07_bt_creator(clip_negative_radiances=clip_negative_radiances)
    clipped_ir = 134.68753 if clip_negative_radiances else np.nan
    expected = np.array(
        [
            clipped_ir,
            304.97037,
            332.22778,
            354.6147,
            374.08688,
            391.58655,
            407.64786,
            422.60635,
            436.68802,
            np.nan,
        ]
    )
    data_np = _get_and_check_array(res, np.float32)
    _check_area(res)
    _check_dims_and_coords(res)
    np.testing.assert_allclose(
        data_np[0, :10], expected, equal_nan=True, atol=1e-04
    )

    # make sure the attributes from the file are in the data array
    assert "scale_factor" not in res.attrs
    assert "_FillValue" not in res.attrs
    assert res.attrs["standard_name"] == "toa_brightness_temperature"
    assert res.attrs["long_name"] == "Brightness Temperature"


def test_vis_calibrate(c01_refl):
    """Test VIS calibration."""
    res = c01_refl
    expected = np.array(
        [
            7.632808,
            15.265616,
            22.898426,
            30.531233,
            38.164043,
            45.796852,
            53.429657,
            61.062466,
            68.695274,
            np.nan,
        ]
    )
    data_np = _get_and_check_array(res, np.float32)
    _check_area(res)
    _check_dims_and_coords(res)
    np.testing.assert_allclose(data_np[0, :10], expected, equal_nan=True)
    assert "scale_factor" not in res.attrs
    assert "_FillValue" not in res.attrs
    assert res.attrs["standard_name"] == "toa_bidirectional_reflectance"
    assert res.attrs["long_name"] == "Bidirectional Reflectance"


def test_raw_calibrate(c01_counts):
    """Test RAW calibration."""
    res = c01_counts

    # We expect the raw data to be unchanged
    _get_and_check_array(res, np.int16)
    _check_area(res)
    _check_dims_and_coords(res)

    # check for the presence of typical attributes
    assert "scale_factor" in res.attrs
    assert "add_offset" in res.attrs
    assert "_FillValue" in res.attrs
    assert "orbital_parameters" in res.attrs
    assert "platform_shortname" in res.attrs
    assert "scene_id" in res.attrs

    # determine if things match their expected values/types.
    assert res.attrs["standard_name"] == "counts"
    assert res.attrs["long_name"] == "Raw Counts"


@mock.patch("satpy.readers.core.abi.xr")
def test_open_dataset(_):  # noqa: PT019
    """Test opening a dataset."""
    openable_thing = mock.MagicMock()

    NC_ABI_L1B(openable_thing, {"platform_shortname": "g16"}, {})
    openable_thing.open.assert_called()



"""Generate the area definition list restructuredtext document.

This should be run once before generating the sphinx documentation to
produce the ``area_def_list.rst`` file referenced by
``satpy/resample/__init__.py``.

"""
import logging
import pathlib
import sys
from datetime import datetime

import bokeh
import geoviews as gv
import geoviews.feature as gf
from bokeh.embed import components
from jinja2 import Template
from pyresample._formatting_html import _load_static_files
from pyresample.area_config import area_repr, load_area
from pyresample.utils.proj4 import ignore_pyproj_proj_warnings
from reader_table import rst_table_header, rst_table_row

from satpy.area import get_area_file

logger = logging.getLogger(__name__)

gv.extension("bokeh")


TEMPLATE = '''

{{ table_header }}
{% for area_name, area_def in areas.items() if area_def._repr_html_ is defined %}
{{ create_table_row(area_name, area_def) }}
{% endfor %}


.. raw:: html

    {{ resources | indent(5) }}
    {{ pyr_icons_svg | indent(5) }}
    <style>
    {{ pyr_css_style | indent(5) }}
    </style>
    {{ script | indent(5)}}

{% for area_name, area_div in area_divs_dict.items() %}

{{ area_name }}
{{ rst_underline('^', area_name|length) }}

.. raw:: html

    {{ area_repr(areas[area_name], map_content=area_div, include_header=False, include_static_files=False) |
       indent(5) }}
    <br>

{% endfor %}
'''  # noqa: Q001


def main():
    """Parse CLI arguments and generate area definition list file."""
    from argparse import ArgumentParser

    parser = ArgumentParser(description="Generate restructuredtext area definition list for sphinx documentation")
    parser.add_argument("--area-file",
                        help="Input area YAML file to read")
    parser.add_argument("-o", "--output-file",
                        type=pathlib.Path,
                        help="HTML or restructuretext filename to create. "
                             "Defaults to 'area_def_list.rst' in the "
                             "documentation source directory.")
    args = parser.parse_args()
    logging.basicConfig(level=logging.INFO)

    if args.output_file is None:
        args.output_file = str(pathlib.Path(__file__).resolve().parent / "area_def_list.rst")
    area_file = args.area_file
    if area_file is None:
        area_file = get_area_file()[0]

    area_list = load_area(area_file)
    areas_dict = {_area_name(area): area for area in area_list}
    logger.info(f"Generating bokeh plots ({datetime.now()})...")
    script, divs_dict = _generate_html_map_divs(areas_dict)
    logger.info(f"Done generating bokeh plots ({datetime.now()})")

    def rst_underline(ch, num_chars):
        return ch * num_chars

    template = Template(TEMPLATE)
    icons_svg, css_style = _load_static_files()
    logger.info(f"Rendering document ({datetime.now()})...")
    res = template.render(
        resources=bokeh.resources.CDN.render(),
        script=script,
        area_divs_dict=divs_dict,
        areas=areas_dict,
        rst_underline=rst_underline,
        area_repr=area_repr,
        pyr_icons_svg=icons_svg,
        pyr_css_style=css_style,
        table_header=rst_table_header("Area Definitions", header=["Name", "Description", "Projection"],
                                      widths="auto", class_name="area-table"),
        create_table_row=_area_table_row,
    )
    logger.info(f"Done rendering document ({datetime.now()})")

    with open(args.output_file, mode="w") as f:
        f.write(res)


def _area_name(area_def) -> str:
    if hasattr(area_def, "attrs"):
        # pyresample 2
        return area_def.attrs["name"]
    # pyresample 1
    return area_def.area_id


def _area_table_row(area_name, area_def):
    with ignore_pyproj_proj_warnings():
        area_proj = area_def.proj_dict.get("proj")
    return rst_table_row([f"`{area_name}`_", area_def.description, area_proj])


def _generate_html_map_divs(areas_dict: dict) -> tuple[str, dict]:
    areas_bokeh_models = {}
    for area_name, area_def in areas_dict.items():
        if not hasattr(area_def, "to_cartopy_crs"):
            logger.info(f"Skipping {area_name} because it can't be converted to cartopy CRS")
            continue
        crs = area_def.to_cartopy_crs()

        features = gv.Overlay([gf.ocean, gf.land, gf.borders, gf.coastline])
        f = gv.render(
            features.opts(
                toolbar=None,
                default_tools=[],
                projection=crs,
                xlim=crs.bounds[:2],
                ylim=crs.bounds[2:],
            ),
            backend="bokeh")
        areas_bokeh_models[area_name] = f

    script, divs_dict = components(areas_bokeh_models)
    return script, divs_dict


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018, 2022 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.ghrsst_l2 module."""

import datetime as dt
import os
import tarfile
from pathlib import Path

import numpy as np
import pytest
import xarray as xr

from satpy.readers.ghrsst_l2 import GHRSSTL2FileHandler


class TestGHRSSTL2Reader:
    """Test Sentinel-3 SST L2 reader."""

    def setup_method(self, tmp_path):
        """Create a fake osisaf ghrsst dataset."""
        self.base_data = np.array(([-32768, 1135, 1125], [1138, 1128, 1080]))
        self.lon_data = np.array(([-13.43, 1.56, 11.25], [-11.38, 1.28, 10.80]))
        self.lat_data = np.array(([43.43, 55.56, 61.25], [41.38, 50.28, 60.80]))
        self.lon = xr.DataArray(
            self.lon_data,
            dims=("nj", "ni"),
            attrs={"standard_name": "longitude",
                   "units": "degrees_east",
                   }
        )
        self.lat = xr.DataArray(
            self.lat_data,
            dims=("nj", "ni"),
            attrs={"standard_name": "latitude",
                   "units": "degrees_north",
                   }
        )
        self.sst = xr.DataArray(
            self.base_data,
            dims=("nj", "ni"),
            attrs={"scale_factor": 0.01, "add_offset": 273.15,
                   "_FillValue": -32768, "units": "kelvin",
                   }
        )
        self.fake_dataset = xr.Dataset(
            data_vars={
                "sea_surface_temperature": self.sst,
                "longitude": self.lon,
                "latitude": self.lat,
            },
            attrs={
                "start_time": "20220321T112640Z",
                "stop_time": "20220321T145711Z",
                "platform": "NOAA20",
                "sensor": "VIIRS",
            },
        )

    def _create_tarfile_with_testdata(self, mypath):
        """Create a 'fake' testdata set in a tar file."""
        slstr_fakename = "S3A_SL_2_WST_MAR_O_NR_003.SEN3"
        tarfile_fakename = "S3A_SL_2_WST_MAR_O_NR_003.SEN3.tar"

        slstrdir = mypath / slstr_fakename
        slstrdir.mkdir(parents=True, exist_ok=True)
        tarfile_path = mypath / tarfile_fakename

        ncfilename = slstrdir / "L2P_GHRSST-SSTskin-202204131200.nc"
        self.fake_dataset.to_netcdf(os.fspath(ncfilename))
        xmlfile_path = slstrdir / "xfdumanifest.xml"
        xmlfile_path.touch()

        with tarfile.open(name=tarfile_path, mode="w") as tar:
            tar.add(os.fspath(ncfilename), arcname=Path(slstr_fakename) / ncfilename.name)
            tar.add(os.fspath(xmlfile_path), arcname=Path(slstr_fakename) / xmlfile_path.name)

        return tarfile_path

    def test_instantiate_single_netcdf_file(self, tmp_path):
        """Test initialization of file handlers - given a single netCDF file."""
        filename_info = {}
        tmp_filepath = tmp_path / "fake_dataset.nc"
        self.fake_dataset.to_netcdf(os.fspath(tmp_filepath))

        GHRSSTL2FileHandler(os.fspath(tmp_filepath), filename_info, None)

    def test_instantiate_tarfile(self, tmp_path):
        """Test initialization of file handlers - given a tar file as in the case of the SAFE format."""
        filename_info = {}
        tarfile_path = self._create_tarfile_with_testdata(tmp_path)

        GHRSSTL2FileHandler(os.fspath(tarfile_path), filename_info, None)

    def test_get_dataset(self, tmp_path):
        """Test retrieval of datasets."""
        filename_info = {}
        tmp_filepath = tmp_path / "fake_dataset.nc"
        self.fake_dataset.to_netcdf(os.fspath(tmp_filepath))

        test = GHRSSTL2FileHandler(os.fspath(tmp_filepath), filename_info, None)

        test.get_dataset("longitude", {"standard_name": "longitude"})
        test.get_dataset("latitude", {"standard_name": "latitude"})
        test.get_dataset("sea_surface_temperature", {"standard_name": "sea_surface_temperature"})

        with pytest.raises(KeyError):
            test.get_dataset("erroneous dataset", {"standard_name": "erroneous dataset"})

    def test_get_sensor(self, tmp_path):
        """Test retrieval of the sensor name from the netCDF file."""
        dt_valid = dt.datetime(2022, 3, 21, 11, 26, 40)  # 202203211200Z
        filename_info = {"field_type": "NARSST", "generating_centre": "FRA_",
                         "satid": "NOAA20_", "valid_time": dt_valid}

        tmp_filepath = tmp_path / "fake_dataset.nc"
        self.fake_dataset.to_netcdf(os.fspath(tmp_filepath))

        test = GHRSSTL2FileHandler(os.fspath(tmp_filepath), filename_info, None)
        assert test.sensor == "viirs"

    def test_get_start_and_end_times(self, tmp_path):
        """Test retrieval of the sensor name from the netCDF file."""
        dt_valid = dt.datetime(2022, 3, 21, 11, 26, 40)  # 202203211200Z
        good_start_time = dt.datetime(2022, 3, 21, 11, 26, 40)  # 20220321T112640Z
        good_stop_time = dt.datetime(2022, 3, 21, 14, 57, 11)  # 20220321T145711Z

        filename_info = {"field_type": "NARSST", "generating_centre": "FRA_",
                         "satid": "NOAA20_", "valid_time": dt_valid}

        tmp_filepath = tmp_path / "fake_dataset.nc"
        self.fake_dataset.to_netcdf(os.fspath(tmp_filepath))

        test = GHRSSTL2FileHandler(os.fspath(tmp_filepath), filename_info, None)

        assert test.start_time == good_start_time
        assert test.end_time == good_stop_time



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""FCI base reader tests package."""

from satpy.readers.core.fci import calculate_area_extent
from satpy.tests.utils import make_dataid


def test_calculate_area_extent():
    """Test function for calculate_area_extent."""
    dataset_id = make_dataid(name="dummy", resolution=2000.0)

    area_dict = {
        "nlines": 5568,
        "ncols": 5568,
        "line_step": dataset_id["resolution"],
        "column_step": dataset_id["resolution"],
    }

    area_extent = calculate_area_extent(area_dict)

    expected = (-5568000.0, 5568000.0, 5568000.0, -5568000.0)

    assert area_extent == expected



"""Native resampler."""

import warnings
from math import lcm  # type: ignore

import dask.array as da
import numpy as np
import xarray as xr
from pyresample.resampler import BaseResampler as PRBaseResampler

from satpy.resample.base import _update_resampled_coords
from satpy.utils import PerformanceWarning, get_legacy_chunk_size

CHUNK_SIZE = get_legacy_chunk_size()


class NativeResampler(PRBaseResampler):
    """Expand or reduce input datasets to be the same shape.

    If data is higher resolution (more pixels) than the destination area
    then data is averaged to match the destination resolution.

    If data is lower resolution (less pixels) than the destination area
    then data is repeated to match the destination resolution.

    This resampler does not perform any caching or masking due to the
    simplicity of the operations.

    """

    def resample(self, data, cache_dir=None, mask_area=False, **kwargs):
        """Run NativeResampler."""
        # use 'mask_area' with a default of False. It wouldn't do anything.
        return super(NativeResampler, self).resample(data,
                                                     cache_dir=cache_dir,
                                                     mask_area=mask_area,
                                                     **kwargs)

    @classmethod
    def _expand_reduce(cls, d_arr, repeats):
        """Expand reduce."""
        d_arr = _ensure_dask_array(d_arr)
        if all(x == 1 for x in repeats.values()):
            return d_arr
        if all(x >= 1 for x in repeats.values()):
            return _replicate(d_arr, repeats)
        if all(x <= 1 for x in repeats.values()):
            # reduce
            y_size = 1. / repeats[0]
            x_size = 1. / repeats[1]
            return _aggregate(d_arr, y_size, x_size)
        raise ValueError("Must either expand or reduce in both "
                         "directions")

    def compute(self, data, expand=True, **kwargs):
        """Resample data with NativeResampler."""
        if isinstance(self.target_geo_def, (list, tuple)):
            # find the highest/lowest area among the provided
            test_func = max if expand else min
            target_geo_def = test_func(self.target_geo_def,
                                       key=lambda x: x.shape)
        else:
            target_geo_def = self.target_geo_def

        # convert xarray backed with numpy array to dask array
        repeats = _get_repeats(target_geo_def, data)

        d_arr = self._expand_reduce(data.data, repeats)
        new_data = xr.DataArray(d_arr, dims=data.dims)
        return _update_resampled_coords(data, new_data, target_geo_def)


def _ensure_dask_array(d_arr):
    if not isinstance(d_arr, da.Array):
        d_arr = da.from_array(d_arr, chunks=CHUNK_SIZE)
    return d_arr


def _get_repeats(target_geo_def, data):
    y_axis, x_axis = _get_axes(data)
    out_shape = target_geo_def.shape
    in_shape = data.shape
    y_repeats = out_shape[0] / float(in_shape[y_axis])
    x_repeats = out_shape[1] / float(in_shape[x_axis])
    repeats = {axis_idx: 1. for axis_idx in range(data.ndim) if axis_idx not in [y_axis, x_axis]}
    repeats[y_axis] = y_repeats
    repeats[x_axis] = x_repeats

    return repeats


def _get_axes(data):
    if "x" not in data.dims or "y" not in data.dims:
        if data.ndim not in [2, 3]:
            raise ValueError("Can only handle 2D or 3D arrays without dimensions.")
        # assume rows is the second to last axis
        y_axis = data.ndim - 2
        x_axis = data.ndim - 1
    else:
        y_axis = data.dims.index("y")
        x_axis = data.dims.index("x")

    return y_axis, x_axis


def _aggregate(d, y_size, x_size):
    """Average every 4 elements (2x2) in a 2D array."""
    if d.ndim != 2:
        # we can't guarantee what blocks we are getting and how
        # it should be reshaped to do the averaging.
        raise ValueError("Can't aggregrate (reduce) data arrays with "
                         "more than 2 dimensions.")
    if not (x_size.is_integer() and y_size.is_integer()):
        raise ValueError("Aggregation factors are not integers")
    y_size = int(y_size)
    x_size = int(x_size)
    d = _rechunk_if_nonfactor_chunks(d, y_size, x_size)
    new_chunks = (tuple(int(x / y_size) for x in d.chunks[0]),
                  tuple(int(x / x_size) for x in d.chunks[1]))
    return da.core.map_blocks(_mean, d, y_size, x_size,
                              meta=np.array((), dtype=d.dtype),
                              dtype=d.dtype, chunks=new_chunks)


def _mean(data, y_size, x_size):
    rows, cols = data.shape
    new_shape = (int(rows / y_size), int(y_size),
                 int(cols / x_size), int(x_size))
    data_mean = np.nanmean(data.reshape(new_shape), axis=(1, 3))
    return data_mean


def _replicate(d_arr, repeats):
    """Repeat data pixels by the per-axis factors specified."""
    repeated_chunks = _get_replicated_chunk_sizes(d_arr, repeats)
    d_arr = d_arr.map_blocks(_repeat_by_factor,
                             meta=np.array((), dtype=d_arr.dtype),
                             dtype=d_arr.dtype,
                             chunks=repeated_chunks)
    return d_arr


def _get_replicated_chunk_sizes(d_arr, repeats):
    repeated_chunks = []
    for axis, axis_chunks in enumerate(d_arr.chunks):
        factor = repeats[axis]
        if not factor.is_integer():
            raise ValueError("Expand factor must be a whole number")
        repeated_chunks.append(tuple(x * int(factor) for x in axis_chunks))
    return tuple(repeated_chunks)


def _repeat_by_factor(data, block_info=None):
    if block_info is None:
        return data
    out_shape = block_info[None]["chunk-shape"]
    out_data = data
    for axis, axis_size in enumerate(out_shape):
        in_size = data.shape[axis]
        out_data = np.repeat(out_data, int(axis_size / in_size), axis=axis)
    return out_data


def _rechunk_if_nonfactor_chunks(dask_arr, y_size, x_size):
    new_chunks = list(dask_arr.chunks)
    need_rechunk = False
    for dim_idx, agg_size in enumerate([y_size, x_size]):
        if dask_arr.shape[dim_idx] % agg_size != 0:
            raise ValueError("Aggregation requires arrays with shapes divisible by the factor.")
        need_rechunk |= _check_chunking(new_chunks, dask_arr, dim_idx, agg_size)
    if need_rechunk:
        warnings.warn(
            "Array chunk size is not divisible by aggregation factor. "
            "Re-chunking to continue native resampling.",
            PerformanceWarning,
            stacklevel=5
        )
        dask_arr = dask_arr.rechunk(tuple(new_chunks))
    return dask_arr


def _check_chunking(new_chunks, dask_arr, dim_idx, agg_size):
    need_rechunk = False
    for chunk_size in dask_arr.chunks[dim_idx]:
        if chunk_size % agg_size != 0:
            need_rechunk = True
            new_dim_chunk = lcm(chunk_size, agg_size)
            new_chunks[dim_idx] = new_dim_chunk
    return need_rechunk


def get_resampler_classes():
    """Get classes based on native resampler."""
    return {"native": NativeResampler}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Vaisala Global Lightning Dataset 360 reader for Universal ASCII Lightning Format 2 (UALF2).

Vaisala Global Lightning Dataset GLD360 is data as a service
that provides real-time lightning data for accurate and early
detection and tracking of severe weather. The data provided is
generated by a Vaisala owned and operated world-wide lightning
detection sensor network.

References:
- [GLD360] https://www.vaisala.com/en/products/data-subscriptions-and-reports/data-sets/gld360
- [SMHI] https://opendata.smhi.se/apidocs/lightning/parameters.html

"""

import logging
from datetime import timedelta

import dask.dataframe as dd
import numpy as np
import xarray as xr

from satpy.readers.core.file_handlers import BaseFileHandler

logger = logging.getLogger(__name__)

UALF2_DTYPES = {
    "ualf_record_type": np.uint8,
    "network_type": np.uint8,
    "year": str,
    "month": str,
    "day": str,
    "hour": str,
    "minute": str,
    "second": str,
    "latitude": np.float32,
    "longitude": np.float32,
    "altitude": np.uint16,
    "altitude_uncertainty": np.uint16,
    "peak_current": np.int16,
    "vhf_range": np.float32,
    "multiplicity_flash": np.uint8,
    "cloud_pulse_count": np.int16,
    "number_of_sensors": np.uint8,
    "degree_freedom_for_location": np.uint8,
    "error_ellipse_angle": np.float32,
    "error_ellipse_max_axis_length": np.float32,
    "error_ellipse_min_axis_length": np.float32,
    "chi_squared_value_location_optimization": np.float32,
    "wave_form_rise_time": np.float32,
    "wave_form_peak_to_zero_time": np.float32,
    "wave_form_max_rate_of_rise": np.float32,
    "cloud_indicator": bool,
    "angle_indicator": bool,
    "signal_indicator": bool,
    "timing_indicator": bool,
}


def _create_column_names():
    """Insert nanoseconds in the column names to a correct index."""
    tmp = [*UALF2_DTYPES]
    idx = tmp.index("second") + 1
    tmp.insert(idx, "nanosecond")

    return tmp


UALF2_COLUMN_NAMES = _create_column_names()


class VaisalaGld360Ualf2FileHandler(BaseFileHandler):
    """FileHandler for Vaisala GLD360 data in UALF2-format."""

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize FileHandler."""
        super(VaisalaGld360Ualf2FileHandler, self).__init__(
            filename, filename_info, filetype_info
        )

        self.data = dd.read_csv(
            filename,
            sep="\t",
            names=UALF2_COLUMN_NAMES,
            dtype=UALF2_DTYPES,
            converters={"nanosecond": self.pad_nanoseconds},
        )

        combined_time = (
            self.data["year"]
            + " "
            + self.data["month"]
            + " "
            + self.data["day"]
            + " "
            + self.data["hour"]
            + " "
            + self.data["minute"]
            + " "
            + self.data["second"]
            + " "
            + self.data["nanosecond"]
        )

        self.data["time"] = dd.to_datetime(combined_time, format="%Y %m %d %H %M %S %f")
        self.data = self.data.drop_duplicates()
        self.data = self.data.sort_values("time")

    @property
    def start_time(self):
        """Return start time."""
        return self.filename_info["start_time"]

    @property
    def end_time(self):
        """Return end time."""
        return self.filename_info["start_time"] + timedelta(hours=1)

    def get_dataset(self, dataset_id, dataset_info):
        """Return the dataset."""
        # create xarray and place along y dimension
        dask_structure = self.data[dataset_id["name"]]
        dask_array = dask_structure.to_dask_array(lengths=dask_structure.compute().shape)
        xarr = xr.DataArray(dask_array, dims=["y"])
        xarr.attrs.update(dataset_info)
        return xarr

    @staticmethod
    def pad_nanoseconds(nanoseconds):
        """Read ns values for less than 0.1s correctly (these are not zero-padded in the input files)."""
        return str(nanoseconds).zfill(9)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""VIIRS Active Fires Tests.

This module implements tests for VIIRS Active Fires NetCDF and ASCII file
readers.

"""

import io
import os
import unittest
from unittest import mock

import dask.dataframe as dd
import numpy as np
import pandas as pd

from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.tests.reader_tests.test_netcdf_utils import FakeNetCDF4FileHandler
from satpy.tests.utils import convert_file_content_to_data_array

DEFAULT_FILE_SHAPE = (1, 100)

DEFAULT_LATLON_FILE_DTYPE = np.float32
DEFAULT_LATLON_FILE_DATA = np.arange(start=43, stop=45, step=0.02,
                                     dtype=DEFAULT_LATLON_FILE_DTYPE).reshape(DEFAULT_FILE_SHAPE)

DEFAULT_DETECTION_FILE_DTYPE = np.uint8
DEFAULT_DETECTION_FILE_DATA = np.arange(start=60, stop=100, step=0.4,
                                        dtype=DEFAULT_DETECTION_FILE_DTYPE).reshape(DEFAULT_FILE_SHAPE)

DEFAULT_M13_FILE_DTYPE = np.float32
DEFAULT_M13_FILE_DATA = np.arange(start=300, stop=340, step=0.4,
                                  dtype=DEFAULT_M13_FILE_DTYPE).reshape(DEFAULT_FILE_SHAPE)

DEFAULT_POWER_FILE_DTYPE = np.float32
DEFAULT_POWER_FILE_DATA = np.arange(start=1, stop=25, step=0.24,
                                    dtype=DEFAULT_POWER_FILE_DTYPE).reshape(DEFAULT_FILE_SHAPE)


class FakeModFiresNetCDF4FileHandler(FakeNetCDF4FileHandler):
    """Swap in CDF4 file handler."""

    def get_test_content(self, filename, filename_info, filename_type):
        """Mimic reader input file content."""
        file_content = {}
        file_content["/attr/data_id"] = "AFMOD"
        file_content["/attr/satellite_name"] = "NPP"
        file_content["/attr/instrument_name"] = "VIIRS"

        file_content["Fire Pixels/FP_latitude"] = DEFAULT_LATLON_FILE_DATA
        file_content["Fire Pixels/FP_longitude"] = DEFAULT_LATLON_FILE_DATA
        file_content["Fire Pixels/FP_power"] = DEFAULT_POWER_FILE_DATA
        file_content["Fire Pixels/FP_T13"] = DEFAULT_M13_FILE_DATA
        file_content["Fire Pixels/FP_T13/attr/units"] = "kelvins"
        file_content["Fire Pixels/FP_confidence"] = DEFAULT_DETECTION_FILE_DATA
        file_content["Fire Pixels/attr/units"] = "none"
        file_content["Fire Pixels/shape"] = DEFAULT_FILE_SHAPE

        attrs = ("FP_latitude", "FP_longitude", "FP_T13", "FP_confidence")
        convert_file_content_to_data_array(
            file_content, attrs=attrs,
            dims=("z", "fakeDim0", "fakeDim1"))
        return file_content


class FakeImgFiresNetCDF4FileHandler(FakeNetCDF4FileHandler):
    """Swap in CDF4 file handler."""

    def get_test_content(self, filename, filename_info, filename_type):
        """Mimic reader input file content."""
        file_content = {}
        file_content["/attr/data_id"] = "AFIMG"
        file_content["/attr/satellite_name"] = "NPP"
        file_content["/attr/instrument_name"] = "VIIRS"

        file_content["Fire Pixels/FP_latitude"] = DEFAULT_LATLON_FILE_DATA
        file_content["Fire Pixels/FP_longitude"] = DEFAULT_LATLON_FILE_DATA
        file_content["Fire Pixels/FP_power"] = DEFAULT_POWER_FILE_DATA
        file_content["Fire Pixels/FP_T4"] = DEFAULT_M13_FILE_DATA
        file_content["Fire Pixels/FP_T4/attr/units"] = "kelvins"
        file_content["Fire Pixels/FP_confidence"] = DEFAULT_DETECTION_FILE_DATA

        attrs = ("FP_latitude", "FP_longitude", "FP_T13", "FP_confidence")
        convert_file_content_to_data_array(
            file_content, attrs=attrs,
            dims=("z", "fakeDim0", "fakeDim1"))
        return file_content


class FakeModFiresTextFileHandler(BaseFileHandler):
    """Fake file handler for text files at moderate resolution."""

    def __init__(self, filename, filename_info, filetype_info, **kwargs):
        """Get fake file content from 'get_test_content'."""
        super(FakeModFiresTextFileHandler, self).__init__(filename, filename_info, filetype_info)
        self.file_content = self.get_test_content()

        platform_key = {"NPP": "Suomi-NPP", "J01": "NOAA-20", "J02": "NOAA-21"}

        self.platform_name = platform_key.get(self.filename_info["satellite_name"].upper(), "unknown")

    def get_test_content(self):
        """Create fake test file content."""
        fake_file = io.StringIO(u"""\n\n\n\n\n\n\n\n\n\n\n\n\n\n
        24.64015007, -107.57017517,  317.38290405,   0.75,   0.75,   40,    4.28618050
        25.90660477, -100.06127167,  331.17962646,   0.75,   0.75,   81,   20.61096764""")

        return dd.from_pandas(pd.read_csv(fake_file, skiprows=15, header=None,
                                          names=["latitude", "longitude",
                                                 "T13", "Along-scan", "Along-track",
                                                 "confidence_pct",
                                                 "power"]), chunksize=1)


class FakeImgFiresTextFileHandler(BaseFileHandler):
    """Fake file handler for text files at image resolution."""

    def __init__(self, filename, filename_info, filetype_info, **kwargs):
        """Get fake file content from 'get_test_content'."""
        super(FakeImgFiresTextFileHandler, self).__init__(filename, filename_info, filetype_info)
        self.file_content = self.get_test_content()

    def get_test_content(self):
        """Create fake test file content."""
        fake_file = io.StringIO(u"""\n\n\n\n\n\n\n\n\n\n\n\n\n\n
        24.64015007, -107.57017517,  317.38290405,   0.75,   0.75,   40,    4.28618050
        25.90660477, -100.06127167,  331.17962646,   0.75,   0.75,   81,   20.61096764""")

        platform_key = {"NPP": "Suomi-NPP", "J01": "NOAA-20", "J02": "NOAA-21"}

        self.platform_name = platform_key.get(self.filename_info["satellite_name"].upper(), "unknown")

        return dd.from_pandas(pd.read_csv(fake_file, skiprows=15, header=None,
                                          names=["latitude", "longitude",
                                                 "T4", "Along-scan", "Along-track",
                                                 "confidence_cat",
                                                 "power"]), chunksize=1)


class TestModVIIRSActiveFiresNetCDF4(unittest.TestCase):
    """Test VIIRS Fires Reader."""

    yaml_file = "viirs_edr_active_fires.yaml"

    def setUp(self):
        """Wrap CDF4 file handler with own fake file handler."""
        from satpy._config import config_search_paths
        from satpy.readers.viirs_edr_active_fires import VIIRSActiveFiresFileHandler
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        self.p = mock.patch.object(VIIRSActiveFiresFileHandler, "__bases__", (FakeModFiresNetCDF4FileHandler,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the CDF4 file handler."""
        self.p.stop()

    def test_init(self):
        """Test basic init with no extra parameters."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "AFMOD_j02_d20180829_t2015451_e2017093_b35434_c20180829210527716708_cspp_dev.nc"
        ])
        assert len(loadables) == 1
        r.create_filehandlers(loadables)
        assert r.file_handlers

    def test_load_dataset(self):
        """Test loading all datasets."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "AFMOD_j02_d20180829_t2015451_e2017093_b35434_c20180829210527716708_cspp_dev.nc"
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(["confidence_pct"])
        assert len(datasets) == 1
        for v in datasets.values():
            assert v.attrs["units"] == "%"
            assert v.attrs["_FillValue"] == 255
            assert np.issubdtype(v.dtype, DEFAULT_DETECTION_FILE_DTYPE)

        datasets = r.load(["T13"])
        assert len(datasets) == 1
        for v in datasets.values():
            assert v.attrs["units"] == "K"

        datasets = r.load(["power"])
        assert len(datasets) == 1
        for v in datasets.values():
            assert v.attrs["units"] == "MW"
            assert v.attrs["platform_name"] == "NOAA-21"
            assert v.attrs["sensor"] == "viirs"


class TestImgVIIRSActiveFiresNetCDF4(unittest.TestCase):
    """Test VIIRS Fires Reader."""

    yaml_file = "viirs_edr_active_fires.yaml"

    def setUp(self):
        """Wrap CDF4 file handler with own fake file handler."""
        from satpy._config import config_search_paths
        from satpy.readers.viirs_edr_active_fires import VIIRSActiveFiresFileHandler
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        self.p = mock.patch.object(VIIRSActiveFiresFileHandler, "__bases__", (FakeImgFiresNetCDF4FileHandler,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the CDF4 file handler."""
        self.p.stop()

    def test_init(self):
        """Test basic init with no extra parameters."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "AFIMG_npp_d20180829_t2015451_e2017093_b35434_c20180829210527716708_cspp_dev.nc"
        ])
        assert len(loadables) == 1
        r.create_filehandlers(loadables)
        assert r.file_handlers

    def test_load_dataset(self):
        """Test loading all datasets."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "AFIMG_npp_d20180829_t2015451_e2017093_b35434_c20180829210527716708_cspp_dev.nc"
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(["confidence_cat"])
        assert len(datasets) == 1
        for v in datasets.values():
            assert v.attrs["units"] == "1"
            assert v.attrs["flag_meanings"] == ["low", "medium", "high"]
            assert v.attrs["flag_values"] == [7, 8, 9]

        datasets = r.load(["T4"])
        assert len(datasets) == 1
        for v in datasets.values():
            assert v.attrs["units"] == "K"

        datasets = r.load(["power"])
        assert len(datasets) == 1
        for v in datasets.values():
            assert v.attrs["units"] == "MW"
            assert v.attrs["platform_name"] == "Suomi-NPP"
            assert v.attrs["sensor"] == "viirs"


@mock.patch("satpy.readers.viirs_edr_active_fires.dd.read_csv")
class TestModVIIRSActiveFiresText(unittest.TestCase):
    """Test VIIRS Fires Reader."""

    yaml_file = "viirs_edr_active_fires.yaml"

    def setUp(self):
        """Wrap file handler with own fake file handler."""
        from satpy._config import config_search_paths
        from satpy.readers.viirs_edr_active_fires import VIIRSActiveFiresTextFileHandler
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        self.p = mock.patch.object(VIIRSActiveFiresTextFileHandler, "__bases__", (FakeModFiresTextFileHandler,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the text file handler."""
        self.p.stop()

    def test_init(self, mock_obj):
        """Test basic init with no extra parameters."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "AFEDR_j01_d20180829_t2015451_e2017093_b35434_c20180829210527716708_cspp_dev.txt"
        ])
        assert len(loadables) == 1
        r.create_filehandlers(loadables)
        assert r.file_handlers

    def test_load_dataset(self, csv_mock):
        """Test loading all datasets."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "AFEDR_j01_d20180829_t2015451_e2017093_b35434_c20180829210527716708_cspp_dev.txt"
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(["confidence_pct"])
        assert len(datasets) == 1
        for v in datasets.values():
            assert v.attrs["units"] == "%"

        datasets = r.load(["T13"])
        assert len(datasets) == 1
        for v in datasets.values():
            assert v.attrs["units"] == "K"

        datasets = r.load(["power"])
        assert len(datasets) == 1
        for v in datasets.values():
            assert v.attrs["units"] == "MW"
            assert v.attrs["platform_name"] == "NOAA-20"
            assert v.attrs["sensor"] == "VIIRS"


@mock.patch("satpy.readers.viirs_edr_active_fires.dd.read_csv")
class TestImgVIIRSActiveFiresText(unittest.TestCase):
    """Test VIIRS Fires Reader."""

    yaml_file = "viirs_edr_active_fires.yaml"

    def setUp(self):
        """Wrap file handler with own fake file handler."""
        from satpy._config import config_search_paths
        from satpy.readers.viirs_edr_active_fires import VIIRSActiveFiresTextFileHandler
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        self.p = mock.patch.object(VIIRSActiveFiresTextFileHandler, "__bases__", (FakeImgFiresTextFileHandler,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the text file handler."""
        self.p.stop()

    def test_init(self, mock_obj):
        """Test basic init with no extra parameters."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "AFIMG_npp_d20180829_t2015451_e2017093_b35434_c20180829210527716708_cspp_dev.txt"
        ])
        assert len(loadables) == 1
        r.create_filehandlers(loadables)
        assert r.file_handlers

    def test_load_dataset(self, mock_obj):
        """Test loading all datasets."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "AFIMG_npp_d20180829_t2015451_e2017093_b35434_c20180829210527716708_cspp_dev.txt"
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(["confidence_cat"])
        assert len(datasets) == 1
        for v in datasets.values():
            assert v.attrs["units"] == "1"
            assert v.attrs["flag_meanings"] == ["low", "medium", "high"]
            assert v.attrs["flag_values"] == [7, 8, 9]

        datasets = r.load(["T4"])
        assert len(datasets) == 1
        for v in datasets.values():
            assert v.attrs["units"] == "K"

        datasets = r.load(["power"])
        assert len(datasets) == 1
        for v in datasets.values():
            assert v.attrs["units"] == "MW"
            assert v.attrs["platform_name"] == "Suomi-NPP"
            assert v.attrs["sensor"] == "VIIRS"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Satpy developers
#
# satpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.

"""EUMETSAT EPS-SG Ice Cloud Imager (ICI) Level 1B products reader.

The format is explained in the
`EPS-SG ICI Level 1B Product Format Specification V3A`_.

This version is applicable for the ici test data released in Jan 2021.

.. _EPS-SG ICI Level 1B Product Format Specification V3A: https://www.eumetsat.int/media/47582

"""

import datetime as dt
import logging
from enum import Enum
from functools import cached_property

import dask.array as da
import numpy as np
import xarray as xr
from geotiepoints.geointerpolator import GeoInterpolator

from satpy.readers.core.netcdf import NetCDF4FileHandler

logger = logging.getLogger(__name__)


# PLANCK COEFFICIENTS FOR CALIBRATION AS DEFINED BY EUMETSAT
C1 = 1.191042e-5  # [mW/(sr·m2·cm-4)]
C2 = 1.4387752  # [K·cm]
# MEAN EARTH RADIUS AS DEFINED BY IUGG
MEAN_EARTH_RADIUS = 6371008.7714  # [m]


class InterpolationType(Enum):
    """Enum for interpolation types."""

    LONLAT = 0
    SOLAR_ANGLES = 1
    OBSERVATION_ANGLES = 2


class IciL1bNCFileHandler(NetCDF4FileHandler):
    """Reader class for ICI L1B products in netCDF format."""

    def __init__(self, filename, filename_info, filetype_info, **kwargs):
        """Read the calibration data and prepare the class for dataset reading."""  # noqa: E501
        super().__init__(
            filename, filename_info, filetype_info, auto_maskandscale=True,
        )
        # Read the variables which are required for the calibration
        measurement = "data/measurement_data"
        self._bt_conversion_a = self[f"{measurement}/bt_conversion_a"].values
        self._bt_conversion_b = self[f"{measurement}/bt_conversion_b"].values
        self._channel_cw = self[f"{measurement}/centre_wavenumber"].values
        self._n_samples = self[measurement].n_samples.size
        self._filetype_info = filetype_info
        self.orthorect = filetype_info.get("orthorect", True)

    @property
    def start_time(self):
        """Get observation start time."""
        try:
            start_time = dt.datetime.strptime(
                self["/attr/sensing_start_time_utc"],
                "%Y%m%d%H%M%S.%f",
            )
        except ValueError:
            start_time = dt.datetime.strptime(
                self["/attr/sensing_start_time_utc"],
                "%Y-%m-%d %H:%M:%S.%f",
            )
        return start_time

    @property
    def end_time(self):
        """Get observation end time."""
        try:
            end_time = dt.datetime.strptime(
                self["/attr/sensing_end_time_utc"],
                "%Y%m%d%H%M%S.%f",
            )
        except ValueError:
            end_time = dt.datetime.strptime(
                self["/attr/sensing_end_time_utc"],
                "%Y-%m-%d %H:%M:%S.%f",
            )
        return end_time

    @property
    def platform_name(self):
        """Return platform name."""
        return self["/attr/spacecraft"]

    @property
    def sensor(self):
        """Return sensor."""
        return self["/attr/instrument"]

    @property
    def ssp_lon(self):
        """Return subsatellite point longitude."""
        # This parameter is not applicable to ICI?
        return None

    @property
    def observation_azimuth(self):
        """Get observation azimuth angles."""
        observation_azimuth, _ = self.observation_azimuth_and_zenith
        return observation_azimuth

    @property
    def observation_zenith(self):
        """Get observation zenith angles."""
        _, observation_zenith = self.observation_azimuth_and_zenith
        return observation_zenith

    @property
    def solar_azimuth(self):
        """Get solar azimuth angles."""
        solar_azimuth, _ = self.solar_azimuth_and_zenith
        return solar_azimuth

    @property
    def solar_zenith(self):
        """Get solar zenith angles."""
        _, solar_zenith = self.solar_azimuth_and_zenith
        return solar_zenith

    @property
    def longitude(self):
        """Get longitude coordinates."""
        longitude, _ = self.longitude_and_latitude
        return longitude

    @property
    def latitude(self):
        """Get latitude coordinates."""
        _, latitude = self.longitude_and_latitude
        return latitude

    @cached_property
    def observation_azimuth_and_zenith(self):
        """Get observation azimuth and zenith angles."""
        return self._interpolate(InterpolationType.OBSERVATION_ANGLES)

    @cached_property
    def solar_azimuth_and_zenith(self):
        """Get solar azimuth and zenith angles."""
        return self._interpolate(InterpolationType.SOLAR_ANGLES)

    @cached_property
    def longitude_and_latitude(self):
        """Get longitude and latitude coordinates."""
        return self._interpolate(InterpolationType.LONLAT)

    @staticmethod
    def _interpolate_geo(
        longitude,
        latitude,
        n_samples,
    ):
        """Perform the interpolation of geographic coordinates from tie points to pixel points.

        Args:
            longitude: xarray DataArray containing the longitude dataset to
                interpolate.
            latitude: xarray DataArray containing the longitude dataset to
                interpolate.
            n_samples: int describing number of samples per scan to interpolate
                onto.

        Returns:
            tuple of arrays containing the interpolate values, all the original
                metadata and the updated dimension names.

        """
        third_dim_name = longitude.dims[2]
        horns = longitude[third_dim_name]
        n_scan = longitude.n_scan
        n_subs = longitude.n_subs
        lons = da.zeros((n_scan.size, n_samples, horns.size))
        lats = da.zeros((n_scan.size, n_samples, horns.size))
        n_subs = np.append(
            np.arange(0, n_samples, np.ceil(n_samples / n_subs.size)),
            n_samples - 1
        ).astype(int)
        for horn in horns.values:
            satint = GeoInterpolator(
                (longitude.values[:, :, horn], latitude.values[:, :, horn]),
                (n_scan.values, n_subs),
                (n_scan.values, np.arange(n_samples)),
            )
            lons_horn, lats_horn = satint.interpolate()
            lons[:, :, horn] = lons_horn
            lats[:, :, horn] = lats_horn
        dims = ["y", "x", third_dim_name]
        lon = xr.DataArray(
            lons,
            attrs=longitude.attrs,
            dims=dims,
            coords={third_dim_name: horns},
        )
        lat = xr.DataArray(
            lats,
            attrs=latitude.attrs,
            dims=dims,
            coords={third_dim_name: horns},
        )
        return lon, lat

    def _interpolate_viewing_angle(
        self,
        azimuth,
        zenith,
        n_samples,
    ):
        """Perform the interpolation of angular coordinates from tie points to pixel points.

        Args:
            azimuth: xarray DataArray containing the azimuth angle dataset to
                interpolate.
            zenith: xarray DataArray containing the zenith angle dataset to
                interpolate.
            n_samples: int describing number of samples per scan to interpolate
                onto.

        Returns:
            tuple of arrays containing the interpolate values, all the original
                metadata and the updated dimension names.

        """
        # interpolate onto spherical coords system with origin at equator
        azimuth, zenith = self._interpolate_geo(azimuth, 90. - zenith, n_samples)
        # transform back such that the origin is at the north pole
        zenith = 90. - zenith
        return azimuth, zenith

    def _interpolate(
        self,
        interpolation_type,
    ):
        """Interpolate from tie points to pixel points."""
        try:
            if interpolation_type is InterpolationType.SOLAR_ANGLES:
                var_key1 = self.filetype_info["solar_azimuth"]
                var_key2 = self.filetype_info["solar_zenith"]
                interp_method = self._interpolate_viewing_angle
            elif interpolation_type is InterpolationType.OBSERVATION_ANGLES:
                var_key1 = self.filetype_info["observation_azimuth"]
                var_key2 = self.filetype_info["observation_zenith"]
                interp_method = self._interpolate_viewing_angle
            else:
                var_key1 = self.filetype_info["longitude"]
                var_key2 = self.filetype_info["latitude"]
                interp_method = self._interpolate_geo
            return interp_method(
                self[var_key1],
                self[var_key2],
                self._n_samples,
            )
        except KeyError:
            logger.warning(f"Datasets for {interpolation_type.name} interpolation not correctly defined in YAML file")  # noqa: E501
        return None, None

    @staticmethod
    def _calibrate_bt(radiance, cw, a, b):
        """Perform the calibration to brightness temperature.

        Args:
            radiance: xarray DataArray or numpy ndarray containing the
                radiance values.
            cw: center wavenumber [cm-1].
            a: temperature coefficient [-].
            b: temperature coefficient [K].

        Returns:
            array containing the calibrated brightness
                temperature values.

        """
        return b + (a * C2 * cw / np.log(1 + C1 * cw ** 3 / radiance))

    def _calibrate(self, variable: xr.DataArray, dataset_info: dict) -> xr.DataArray:
        """Perform the calibration.

        Args:
            variable: xarray DataArray containing the dataset to calibrate.
            dataset_info: dictionary of information about the dataset.

        Returns:
            array containing the calibrated values and all the
                original metadata.

        """
        calibration_name = dataset_info["calibration"]
        if calibration_name == "brightness_temperature":
            chan_index = dataset_info["chan_index"]
            cw = self._channel_cw[chan_index]
            a = self._bt_conversion_a[chan_index]
            b = self._bt_conversion_b[chan_index]
            calibrated_variable = self._calibrate_bt(variable, cw, a, b)
            calibrated_variable.attrs = variable.attrs
        elif calibration_name == "radiance":
            calibrated_variable = variable
        else:
            raise ValueError("Unknown calibration %s for dataset %s" % (calibration_name, dataset_info["name"]))  # noqa: E501

        return calibrated_variable

    def _orthorectify(self, variable: xr.DataArray, orthorect_data_name: str) -> xr.DataArray:
        """Perform the orthorectification.

        Args:
            variable: xarray DataArray containing the dataset to correct for
                orthorectification.
            orthorect_data_name: name of the orthorectification correction data
                in the product.

        Returns:
            array containing the corrected values and all the
                original metadata.

        """
        try:
            # Convert the orthorectification delta values from meters to
            # degrees based on the simplified formula using mean Earth radius
            orthorect_data = self[orthorect_data_name]
            dim = self._get_third_dimension_name(orthorect_data)
            orthorect_data = orthorect_data.sel({dim: variable[dim]})
            variable += np.degrees(orthorect_data.values / MEAN_EARTH_RADIUS)
        except KeyError:
            logger.warning("Required dataset %s for orthorectification not available, skipping", orthorect_data_name)  # noqa: E501
        return variable

    @staticmethod
    def _standardize_dims(variable):
        """Standardize dims to y, x."""
        if "n_scan" in variable.dims:
            variable = variable.rename({"n_scan": "y"})
        if "n_samples" in variable.dims:
            variable = variable.rename({"n_samples": "x"})
        if variable.dims[0] == "x":
            variable = variable.transpose("y", "x")
        return variable

    def _filter_variable(self, variable, dataset_info):
        """Filter variable in the third dimension."""
        dim = self._get_third_dimension_name(variable)
        if dim is not None and dim in dataset_info:
            variable = variable.sel({dim: dataset_info[dim]})
        return variable

    @staticmethod
    def _drop_coords(variable):
        """Drop coords that are not in dims."""
        for coord in variable.coords:
            if coord not in variable.dims:
                variable = variable.drop_vars(coord)
        return variable

    @staticmethod
    def _get_third_dimension_name(variable):
        """Get name of the third dimension of the variable."""
        dims = variable.dims
        if len(dims) < 3:
            return None
        return dims[2]

    def _fetch_variable(self, var_key):
        """Fetch variable."""
        if var_key in [
            "longitude",
            "latitude",
            "observation_zenith",
            "observation_azimuth",
            "solar_zenith",
            "solar_azimuth",
        ] and getattr(self, var_key) is not None:
            variable = getattr(self, var_key).copy()
        else:
            variable = self[var_key]
        return variable

    def get_dataset(self, dataset_id, dataset_info):
        """Get dataset using file_key in dataset_info."""
        var_key = dataset_info["file_key"]
        logger.debug(f"Reading in file to get dataset with key {var_key}.")
        try:
            variable = self._fetch_variable(var_key)
        except KeyError:
            logger.warning(f"Could not find key {var_key} in NetCDF file, no valid Dataset created")  # noqa: E501
            return None
        variable = self._filter_variable(variable, dataset_info)
        if dataset_info.get("calibration") is not None:
            variable = self._calibrate(variable, dataset_info)
        if self.orthorect:
            orthorect_data_name = dataset_info.get("orthorect_data", None)
            if orthorect_data_name is not None:
                variable = self._orthorectify(variable, orthorect_data_name)
        variable = self._manage_attributes(variable, dataset_info)
        variable = self._drop_coords(variable)
        variable = self._standardize_dims(variable)
        return variable

    def _manage_attributes(self, variable, dataset_info):
        """Manage attributes of the dataset."""
        variable.attrs.setdefault("units", None)
        variable.attrs.update(dataset_info)
        variable.attrs.update(self._get_global_attributes())
        return variable

    def _get_global_attributes(self):
        """Create a dictionary of global attributes."""
        return {
            "filename": self.filename,
            "start_time": self.start_time,
            "end_time": self.end_time,
            "spacecraft_name": self.platform_name,
            "ssp_lon": self.ssp_lon,
            "sensor": self.sensor,
            "filename_start_time": self.filename_info["sensing_start_time"],
            "filename_end_time": self.filename_info["sensing_end_time"],
            "platform_name": self.platform_name,
            "quality_group": self._get_quality_attributes(),
        }

    def _get_quality_attributes(self):
        """Get quality attributes."""
        quality_group = self["quality"]
        quality_dict = {}
        for key in quality_group:
            # Add the values (as Numpy array) of each variable in the group
            # where possible
            try:
                quality_dict[key] = quality_group[key].values
            except ValueError:
                quality_dict[key] = None
        # Add the attributes of the quality group
        quality_dict.update(quality_group.attrs)
        return quality_dict
